# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the Join and Leave commands."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.commands.eml_who import Who
from mailman.email.message import Message
from mailman.interfaces.command import ContinueProcessing
from mailman.interfaces.member import DeliveryMode, DeliveryStatus, MemberRole
from mailman.model.roster import RosterVisibility
from mailman.runners.command import Results
from mailman.testing.helpers import subscribe
from mailman.testing.layers import ConfigLayer


class TestWho(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        self._command = Who()
        self._member = subscribe(self._mlist, 'Cate')
        subscribe(self._mlist, 'Doug')
        subscribe(self._mlist, 'Anne')
        bmember = subscribe(self._mlist, 'Bart')
        emember = subscribe(self._mlist, 'Elly')
        subscribe(self._mlist, 'Fred')
        bmember.preferences.delivery_status = DeliveryStatus.by_moderator
        emember.preferences.delivery_mode = DeliveryMode.mime_digests
        self._moderator = subscribe(self._mlist, 'Greg',
                                    role=MemberRole.moderator)
        self._owner = subscribe(self._mlist, 'Irma', role=MemberRole.owner)
        self._expected = """\
The results of your email command are provided below.

Members of the ant@example.com mailing list:
    Anne Person <aperson@example.com>
    Bart Person <bperson@example.com>
    Cate Person <cperson@example.com>
    Doug Person <dperson@example.com>
    Elly Person <eperson@example.com>
    Fred Person <fperson@example.com>
"""
        self._noperm = """\
The results of your email command are provided below.

You are not authorized to see the membership list.
"""

    def test_public_roster(self):
        self._mlist.member_roster_visibility = RosterVisibility.public
        # An unknown address should be able to get a public roster.
        msg = Message()
        msg['From'] = 'unknown@example.com'
        results = Results()
        ret = self._command.process(self._mlist, msg, {}, (), results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(self._expected, results._output.getvalue())

    def test_non_public_roster(self):
        self._mlist.member_roster_visibility = RosterVisibility.members
        # An unknown address should not be able to get a non-public roster.
        msg = Message()
        msg['From'] = 'unknown@example.com'
        results = Results()
        ret = self._command.process(self._mlist, msg, {}, (), results)
        self.assertIs(ContinueProcessing.no, ret)
        self.assertEqual(self._noperm, results._output.getvalue())

    def test_member_visible_roster(self):
        self._mlist.member_roster_visibility = RosterVisibility.members
        # A member should be able to get a member visible roster.
        msg = Message()
        msg['From'] = self._member.address.email
        results = Results()
        ret = self._command.process(self._mlist, msg, {}, (), results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(self._expected, results._output.getvalue())

    def test_member_visible_roster_owner(self):
        self._mlist.member_roster_visibility = RosterVisibility.members
        # An owner should be able to get a member visible roster.
        msg = Message()
        msg['From'] = self._owner.address.email
        results = Results()
        ret = self._command.process(self._mlist, msg, {}, (), results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(self._expected, results._output.getvalue())

    def test_member_visible_roster_moderator(self):
        self._mlist.member_roster_visibility = RosterVisibility.members
        # A moderator should be able to get a member visible roster.
        msg = Message()
        msg['From'] = self._moderator.address.email
        results = Results()
        ret = self._command.process(self._mlist, msg, {}, (), results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(self._expected, results._output.getvalue())

    def test_moderator_visible_roster(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # A member should not be able to get a moderator visible roster.
        msg = Message()
        msg['From'] = self._member.address.email
        results = Results()
        ret = self._command.process(self._mlist, msg, {}, (), results)
        self.assertIs(ContinueProcessing.no, ret)
        self.assertEqual(self._noperm, results._output.getvalue())

    def test_moderator_visible_roster_nonmember(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # An unknown address can't get a moderator visible roster.
        msg = Message()
        msg['From'] = 'unknown@example.com'
        results = Results()
        ret = self._command.process(self._mlist, msg, {}, (), results)
        self.assertIs(ContinueProcessing.no, ret)
        self.assertEqual(self._noperm, results._output.getvalue())

    def test_moderaror_visible_roster_owner(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # An owner should be able to get a moderator visible roster.
        msg = Message()
        msg['From'] = self._owner.address.email
        results = Results()
        ret = self._command.process(self._mlist, msg, {}, (), results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(self._expected, results._output.getvalue())

    def test_moderaror_visible_roster_moderator(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # A moderator should be able to get a moderator visible roster.
        msg = Message()
        msg['From'] = self._moderator.address.email
        results = Results()
        ret = self._command.process(self._mlist, msg, {}, (), results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(self._expected, results._output.getvalue())

    def test_args_delivery_no(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # Test delivery=disabled.
        msg = Message()
        msg['From'] = self._moderator.address.email
        results = Results()
        args = ['delivery=disabled']
        ret = self._command.process(self._mlist, msg, {}, args, results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(results._output.getvalue(), """\
The results of your email command are provided below.

Members of the ant@example.com mailing list:
    Bart Person <bperson@example.com>
""")

    def test_args_delivery_yes(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # Test delivery=enabled.
        msg = Message()
        msg['From'] = self._moderator.address.email
        results = Results()
        args = ['delivery=enabled']
        ret = self._command.process(self._mlist, msg, {}, args, results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(results._output.getvalue(), """\
The results of your email command are provided below.

Members of the ant@example.com mailing list:
    Anne Person <aperson@example.com>
    Cate Person <cperson@example.com>
    Doug Person <dperson@example.com>
    Elly Person <eperson@example.com>
    Fred Person <fperson@example.com>
""")

    def test_args_delivery_bogus(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # Test delivery=bogus.
        msg = Message()
        msg['From'] = self._moderator.address.email
        results = Results()
        args = ['delivery=bogus']
        ret = self._command.process(self._mlist, msg, {}, args, results)
        self.assertIs(ContinueProcessing.no, ret)
        self.assertEqual(results._output.getvalue(), """\
The results of your email command are provided below.

Unrecognized or invalid argument(s):
delivery=bogus
""")

    def test_args_mode_yes(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # Test mode=digest.
        msg = Message()
        msg['From'] = self._moderator.address.email
        results = Results()
        args = ['mode=digest']
        ret = self._command.process(self._mlist, msg, {}, args, results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(results._output.getvalue(), """\
The results of your email command are provided below.

Members of the ant@example.com mailing list:
    Elly Person <eperson@example.com>
""")

    def test_args_mode_no(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # Test mode=regular.
        msg = Message()
        msg['From'] = self._moderator.address.email
        results = Results()
        args = ['mode=regular']
        ret = self._command.process(self._mlist, msg, {}, args, results)
        self.assertIs(ContinueProcessing.yes, ret)
        self.assertEqual(results._output.getvalue(), """\
The results of your email command are provided below.

Members of the ant@example.com mailing list:
    Anne Person <aperson@example.com>
    Bart Person <bperson@example.com>
    Cate Person <cperson@example.com>
    Doug Person <dperson@example.com>
    Fred Person <fperson@example.com>
""")

    def test_args_mode_bogus(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # Test mode=bogus.
        msg = Message()
        msg['From'] = self._moderator.address.email
        results = Results()
        args = ['mode=bogus']
        ret = self._command.process(self._mlist, msg, {}, args, results)
        self.assertIs(ContinueProcessing.no, ret)
        self.assertEqual(results._output.getvalue(), """\
The results of your email command are provided below.

Unrecognized or invalid argument(s):
mode=bogus
""")

    def test_args_multi_bogus(self):
        self._mlist.member_roster_visibility = RosterVisibility.moderators
        # Test multiple args, some bogus.
        msg = Message()
        msg['From'] = self._moderator.address.email
        results = Results()
        args = ['mode=regular', 'digest=bogus', 'mode=bogus', 'delivery=none']
        ret = self._command.process(self._mlist, msg, {}, args, results)
        self.assertIs(ContinueProcessing.no, ret)
        self.assertEqual(results._output.getvalue(), """\
The results of your email command are provided below.

Unrecognized or invalid argument(s):
digest=bogus
mode=bogus
delivery=none
""")



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Datetime utilities.

Use these functions to produce variable times rather than the built-in
datetime.datetime.now() and datetime.date.today().  These are better
instrumented for testing purposes.
"""

import datetime

from mailman.testing import layers
from public import public


# Python always sets the locale to 'C' locale unless the user explicitly calls
# locale.setlocale(locale.LC_ALL, '').  Since we never do this in Mailman (and
# no library better do it either!) this will safely give us expected RFC 5322
# Date headers.
public(RFC822_DATE_FMT='%a, %d %b %Y %H:%M:%S %z')


# Definition of UTC timezone, taken from
# https://docs.python.org/3/library/datetime.html
ZERO = datetime.timedelta(0)


@public
class UTC(datetime.tzinfo):
    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return 'UTC'

    def dst(self, dt):
        return ZERO


utc = UTC()
public(utc=utc)
_missing = object()


class DateFactory:
    """A factory for today() and now() that works with testing."""

    # The predictable time.
    predictable_now = None
    predictable_today = None

    def now(self, tz=_missing, strip_tzinfo=True):
        # We can't automatically fast-forward because some tests require us to
        # stay on the same day for a while, e.g. autorespond.txt.
        if tz is _missing:
            tz = utc
        # Storm cannot yet handle datetimes with tz suffixes.  Assume we're
        # using UTC datetimes everywhere, so set the tzinfo to None.  This
        # does *not* change the actual time values.  LP: #280708
        tz_now = (self.predictable_now
                  if layers.is_testing()
                  else datetime.datetime.now(tz))
        return (tz_now.replace(tzinfo=None)
                if strip_tzinfo
                else tz_now)

    def today(self):
        return (self.predictable_today
                if layers.is_testing()
                else datetime.date.today())

    @classmethod
    def reset(cls):
        cls.predictable_now = datetime.datetime(2005, 8, 1, 7, 49, 23,
                                                tzinfo=utc)
        cls.predictable_today = cls.predictable_now.date()

    @classmethod
    def fast_forward(cls, days=1):
        cls.predictable_now += datetime.timedelta(days=days)
        cls.predictable_today = cls.predictable_now.date()


factory = DateFactory()
public(factory=factory)
factory.reset()

public(today=factory.today)
public(now=factory.now)
layers.MockAndMonkeyLayer.register_reset(factory.reset)



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interface to bounce detection components."""

from enum import Enum
from public import public
from zope.interface import Attribute, Interface


class InvalidBounceEvent(Exception):
    """An invalid bounce event was found during processing.

    This either means the email is no longer a member or the MailingList does
    not exist anymore.
    """


@public
class BounceContext(Enum):
    """The context in which the bounce was detected."""

    # This is a normal bounce detection.  IOW, Mailman received a bounce in
    # response to a mailing list post.
    normal = 1

    # A probe email bounced.  This can be considered a bit more serious, since
    # it occurred in response to a specific message to a specific user.
    probe = 2


@public
class UnrecognizedBounceDisposition(Enum):
    # Just throw the message away.
    discard = 0
    # Forward the message to the list administrators, which includes both the
    # owners and the moderators.
    administrators = 1
    # Forward the message to the site owner.
    site_owner = 2


@public
class IBounceEvent(Interface):
    """Registration record for a single bounce event."""

    list_id = Attribute(
        """The List-ID of the mailing list that received this bounce.""")

    email = Attribute(
        """The email address that bounced.""")

    timestamp = Attribute(
        """The timestamp for when the bounce was received.""")

    message_id = Attribute(
        """The Message-ID of the bounce message.""")

    context = Attribute(
        """Where was the bounce detected?""")

    processed = Attribute(
        """Has this bounce event been processed?""")


@public
class IBounceProcessor(Interface):
    """Manager/processor of bounce events."""

    def register(mlist, email, msg, context=None):
        """Register a bounce event.

        :param mlist: The mailing list that the bounce occurred on.
        :type mlist: IMailingList
        :param email: The email address that is bouncing.
        :type email: str
        :param msg: The bounce message.
        :type msg: email.message.Message
        :param context: In what context was the bounce detected?  The default
            is 'normal' context (i.e. we received a normal bounce for the
            address).
        :type context: BounceContext
        :return: The registered bounce event.
        :rtype: IBounceEvent
        """

    def process_event(event):
        """Process a single bounce event.

        :param event: The bounce event to process.
        :type event: IBounceEvent
        """

    def send_warnings_and_remove():
        """Send warnings to disabled users and remove them if needed.

        All users, that have their delivery disabled, get warnings about being
        un-subscribed until Mailinglist's maximum configured numbered of
        warnings are sent and then remove the user.
        """

    events = Attribute(
        """An iterator over all events.""")

    unprocessed = Attribute(
        """An iterator over all unprocessed bounce events.""")



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Global events."""

from mailman.app import domain, membership, moderator, subscriptions
from mailman.core import i18n, switchboard
from mailman.languages import manager as language_manager
from mailman.styles import manager as style_manager
from mailman.utilities import passwords
from public import public
from zope import event


@public
def initialize():
    """Initialize global event subscribers."""
    event.subscribers.extend([
        domain.handle_DomainDeletingEvent,
        i18n.handle_ConfigurationUpdatedEvent,
        language_manager.handle_ConfigurationUpdatedEvent,
        membership.handle_SubscriptionEvent,
        moderator.handle_ListDeletingEvent,
        passwords.handle_ConfigurationUpdatedEvent,
        style_manager.handle_ConfigurationUpdatedEvent,
        subscriptions.handle_ListDeletingEvent,
        subscriptions.handle_SubscriptionConfirmationNeededEvent,
        subscriptions.handle_SubscriptionInvitationNeededEvent,
        subscriptions.handle_UnsubscriptionConfirmationNeededEvent,
        switchboard.handle_ConfigurationUpdatedEvent,
        ])



# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test various preference functionality."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.database.transaction import transaction
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import call_api
from mailman.testing.layers import RESTLayer
from urllib.error import HTTPError
from zope.component import getUtility


class TestPreferences(unittest.TestCase):
    """Test various preference functionality."""

    layer = RESTLayer

    def setUp(self):
        user_manager = getUtility(IUserManager)
        with transaction():
            self._mlist = create_list('test@example.com')
            anne = user_manager.create_address('anne@example.com')
            self._member = self._mlist.subscribe(anne)

    def test_read_only_member_all_preferences(self):
        # A member's combined preferences are read-only.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/members/1/all/preferences', {
                     'delivery_status': 'enabled',
                     }, method='PATCH')
        # The resource at this endpoint doesn't even have a PATCH method.
        self.assertEqual(cm.exception.code, 405)

    def test_read_only_system_preferences(self):
        # The system preferences are read-only.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/preferences', {
                     'delivery_status': 'enabled',
                     }, method='PATCH')
        # The resource at this endpoint doesn't even have a PATCH method.
        self.assertEqual(cm.exception.code, 405)



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The implicit destination rule."""

import re

from contextlib import suppress
from email.utils import getaddresses
from mailman.core.i18n import _
from mailman.interfaces.mailinglist import IAcceptableAliasSet
from mailman.interfaces.rules import IRule
from public import public
from zope.interface import implementer


@public
@implementer(IRule)
class ImplicitDestination:
    """The implicit destination rule."""

    name = 'implicit-dest'
    description = _('Catch messages with implicit destination.')
    record = True

    def check(self, mlist, msg, msgdata):
        """See `IRule`."""
        # Implicit destination checking must be enabled in the mailing list.
        if not mlist.require_explicit_destination:
            return False
        # Messages gated from NNTP will always have an implicit destination so
        # are never checked.
        if msgdata.get('fromusenet'):
            return False
        # Calculate the list of acceptable aliases.  If the alias starts with
        # a caret (i.e. ^), then it's a regular expression to match against.
        aliases = set()
        alias_patterns = set()
        # Adapt the mailing list to the appropriate interface.
        alias_set = IAcceptableAliasSet(mlist)
        for alias in alias_set.aliases:
            if alias.startswith('^'):
                alias_patterns.add(alias)
            else:
                aliases.add(alias)
        # Add the list's posting address, i.e. the explicit address, to the
        # set of acceptable aliases.
        aliases.add(mlist.posting_address)
        # Look at all the recipients.  If the recipient is any acceptable
        # alias (or the explicit posting address), then this rule does not
        # match.  If not, then add it to the set of recipients we'll check
        # against the alias patterns later.
        recipients = set()
        for header in ('to', 'cc', 'resent-to', 'resent-cc'):
            for fullname, address in getaddresses(msg.get_all(header, [])):
                if isinstance(address, bytes):
                    address = address.decode('ascii')
                address = address.lower()
                if address in aliases:
                    return False
                recipients.add(address)
        # Now for all alias patterns, see if any of the recipients matches a
        # pattern.  If so, then this rule does not match.
        for pattern in alias_patterns:
            escaped = re.escape(pattern)
            for recipient in recipients:
                try:
                    if re.match(pattern, recipient, re.IGNORECASE):
                        return False
                except re.error:
                    # The pattern is a malformed regular expression.  Try
                    # matching again with the pattern escaped.
                    with suppress(re.error):
                        if re.match(escaped, recipient, re.IGNORECASE):
                            return False
        # Nothing matched.
        msgdata['moderation_sender'] = msg.sender
        with _.defer_translation():
            # This will be translated at the point of use.
            msgdata.setdefault('moderation_reasons', []).append(
                _('Message has implicit destination'))
        return True



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test configuration file searching."""

import os
import sys
import shutil
import tempfile
import unittest

from contextlib import contextmanager
from mailman.core.initialize import search_for_configuration_file
from mailman.testing.helpers import hackenv


# Here are a couple of context managers that make our tests easier to read.
@contextmanager
def fakedirs(path):
    """Create and clean up a directory hierarchy."""
    os.makedirs(path)
    try:
        yield
    finally:
        shutil.rmtree(path)


@contextmanager
def chdir(new_cwd):
    """Change to the directory, then back again."""
    old_cwd = os.getcwd()
    os.chdir(new_cwd)
    try:
        yield
    finally:
        os.chdir(old_cwd)


@contextmanager
def argv0(new_argv0):
    """Change argv0, then back again."""
    old_argv0 = sys.argv[0]
    sys.argv[0] = new_argv0
    try:
        yield
    finally:
        sys.argv[0] = old_argv0


class TestConfigFileBase(unittest.TestCase):
    """Common test infrastructure."""

    def setUp(self):
        self._root = tempfile.mkdtemp()
        # Ensure that the environment can't cause test failures.
        self.mailman_config_file = os.environ.get('MAILMAN_CONFIG_FILE')
        if self.mailman_config_file is not None:
            del os.environ['MAILMAN_CONFIG_FILE']

    def tearDown(self):
        shutil.rmtree(self._root)
        # Restore the environment, though I'm not actually sure this is
        # necessary.
        if self.mailman_config_file is not None:
            os.environ['MAILMAN_CONFIG_FILE'] = self.mailman_config_file

    def _make_fake(self, path):
        if path.startswith('/'):
            path = path[1:]
        return os.path.join(self._root, path)


class TestConfigFileSearch(TestConfigFileBase):
    """Test various aspects of searching for configuration files.

    Note that the command line -C option is not tested here.
    """

    def test_current_working_directory(self):
        fake_cwd = '/home/alex/mailman/hacking'
        fake_testdir = self._make_fake(fake_cwd)
        config_file = os.path.realpath(
            os.path.join(fake_testdir, 'mailman.cfg'))
        with fakedirs(fake_testdir):
            # Write a mostly empty configuration file.
            with open(os.path.join(fake_testdir, 'mailman.cfg'), 'w') as fp:
                print('# Fake mailman.cfg file', file=fp)
            with chdir(fake_testdir):
                # Sometimes symlinks bite us (eg. OS X /var -> /private/var).
                found = os.path.realpath(search_for_configuration_file())
                self.assertEqual(found, config_file)


class TestConfigFileSearchWithChroot(TestConfigFileBase):
    """Like `TestConfigFileSearch` but with a special os.path.exists()."""

    def setUp(self):
        TestConfigFileBase.setUp(self)
        # We can't actually call os.chroot() unless we're root.  Neither can
        # we write to say /etc/mailman.cfg without being root (of course we
        # wouldn't want to even if we could).  The easiest way to fake a file
        # system that we can write to and test is to hack os.path.exists() to
        # prepend a temporary directory onto the path it tests.
        self._os_path_exists = os.path.exists
        def exists(path):                                  # noqa: E306
            # Strip off the leading slash, otherwise we'll end up with path.
            return self._os_path_exists(self._make_fake(path))
        os.path.exists = exists

    def tearDown(self):
        os.path.exists = self._os_path_exists
        TestConfigFileBase.tearDown(self)

    def test_baseline(self):
        # With nothing set, and no configuration files, just return None.
        self.assertEqual(search_for_configuration_file(), None)

    def test_environment_variable_to_missing_path(self):
        # Test that $MAILMAN_CONFIG_FILE pointing to a non-existent path still
        # returns None.
        with hackenv('MAILMAN_CONFIG_FILE', '/does/not/exist'):
            self.assertEqual(search_for_configuration_file(), None)

    def test_environment_variable(self):
        # Test that $MAILMAN_CONFIG_FILE pointing to an existing path returns
        # that path.
        fake_home = '/home/geddy/testing/mailman'
        fake_testdir = self._make_fake(fake_home)
        config_file = os.path.join(fake_home, 'mailman.cfg')
        with fakedirs(fake_testdir):
            # Write a mostly empty configuration file.
            with open(os.path.join(fake_testdir, 'mailman.cfg'), 'w') as fp:
                print('# Fake mailman.cfg file', file=fp)
            with hackenv('MAILMAN_CONFIG_FILE', config_file):
                self.assertEqual(search_for_configuration_file(), config_file)

    def test_home_dot_file(self):
        # Test ~/.mailman.cfg
        fake_home = '/home/neil'
        fake_testdir = self._make_fake(fake_home)
        config_file = os.path.join(fake_home, '.mailman.cfg')
        with fakedirs(fake_testdir):
            # Write a mostly empty configuration file.
            with open(os.path.join(fake_testdir, '.mailman.cfg'), 'w') as fp:
                print('# Fake mailman.cfg file', file=fp)
            with hackenv('HOME', '/home/neil'):
                self.assertEqual(search_for_configuration_file(), config_file)

    def test_etc_file(self):
        # Test /etc/mailman.cfg
        fake_etc = '/etc'
        fake_testdir = self._make_fake(fake_etc)
        config_file = os.path.join(fake_etc, 'mailman.cfg')
        with fakedirs(fake_testdir):
            # Write a mostly empty configuration file.
            with open(os.path.join(fake_testdir, 'mailman.cfg'), 'w') as fp:
                print('# Fake mailman.cfg file', file=fp)
            self.assertEqual(search_for_configuration_file(), config_file)

    def test_etc_mailman3_file(self):
        # Test /etc/mailman3/mailman.cfg
        fake_etc = '/etc/mailman3'
        fake_testdir = self._make_fake(fake_etc)
        config_file = os.path.join(fake_etc, 'mailman.cfg')
        with fakedirs(fake_testdir):
            # Write a mostly empty configuration file.
            with open(os.path.join(fake_testdir, 'mailman.cfg'), 'w') as fp:
                print('# Fake mailman.cfg file', file=fp)
            self.assertEqual(search_for_configuration_file(), config_file)

    def test_sibling_directory(self):
        # Test $argv0/../../etc/mailman.cfg
        fake_root = '/usr/local/mm3'
        fake_testdir = self._make_fake(fake_root)
        config_file = os.path.join(fake_testdir, 'etc', 'mailman.cfg')
        fake_config_file = os.path.join(fake_root, 'etc', 'mailman.cfg')
        fake_argv0 = os.path.join(fake_root, 'bin', 'mailman')
        with fakedirs(fake_testdir):
            with argv0(fake_argv0):
                os.mkdir(os.path.dirname(config_file))
                # Write a mostly empty configuration file.
                with open(config_file, 'w') as fp:
                    print('# Fake mailman.cfg file', file=fp)
                self.assertEqual(search_for_configuration_file(),
                                 fake_config_file)



# Copyright (C) 2002-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The email commands 'join' and 'subscribe'."""

from email.header import decode_header, make_header
from email.utils import formataddr, parseaddr
from mailman.core.i18n import _
from mailman.interfaces.address import InvalidEmailAddressError
from mailman.interfaces.command import ContinueProcessing, IEmailCommand
from mailman.interfaces.member import (
    AlreadySubscribedError,
    DeliveryMode,
    MembershipIsBannedError,
)
from mailman.interfaces.subscriptions import (
    ISubscriptionManager,
    SubscriptionPendingError,
)
from mailman.interfaces.usermanager import IUserManager
from public import public
from zope.component import getUtility
from zope.interface import implementer


def match_subscriber(email, display_name):
    # Return something matching the email which should be used as the
    # subscriber by the ISubscriptionManager interface.
    manager = getUtility(IUserManager)
    # Is there a user with a preferred address matching the email?
    user = manager.get_user(email)
    if user is not None:
        preferred = user.preferred_address
        if preferred is not None and preferred.email == email.lower():
            return user
    # Is there an address matching the email?
    address = manager.get_address(email)
    if address is not None:
        return address
    # Make a new user and subscribe their first (and only) address.  We can't
    # make the first address their preferred address because it hasn't been
    # verified yet.
    user = manager.make_user(email, display_name)
    return list(user.addresses)[0]


@public
@implementer(IEmailCommand)
class Join:
    """The email 'join' command."""

    name = 'join'
    # XXX 2012-02-29 BAW: DeliveryMode.summary is not yet supported.
    argument_description = ('[digest=<no|mime|plain>] '
                            '[address=user@example.com]')
    description = _("""\
You will be asked to confirm your subscription request and you may be issued a
provisional password.

By using the 'digest' option, you can specify whether you want digest delivery
or not.  If not specified, the mailing list's default delivery mode will be
used.  You can use the 'address' option to request subscription of an address
other than the sender of the command.
""")
    short_description = _('Join this mailing list.')

    def process(self, mlist, msg, msgdata, arguments, results):
        """See `IEmailCommand`."""
        # Parse the arguments.
        delivery_mode, address = self._parse_arguments(arguments, results)
        if delivery_mode is ContinueProcessing.no:
            return ContinueProcessing.no
        if not address:
            # RFC 2047 decode the From: header.
            if msg['from'] is None:
                display_name, email = ('', '')
            else:
                # Parse before decoding in case decoded display_name contains
                # a comma.
                display_name, email = parseaddr(msg['from'])
                display_name = str(make_header(decode_header(display_name)))
        else:
            display_name, email = ('', address)
        # Address could be None or the empty string.
        if not email:
            email = msg.sender
        if not email:
            print(_('${self.name}: No valid address found to subscribe'),
                  file=results)
            return ContinueProcessing.no
        if isinstance(email, bytes):
            email = email.decode('ascii')
        # Have we already seen one join request from this user during the
        # processing of this email?
        joins = getattr(results, 'joins', set())
        if email in joins:
            # Do not register this join.
            return ContinueProcessing.yes
        joins.add(email)
        results.joins = joins
        person = formataddr((display_name, email))                # noqa: F841
        try:
            subscriber = match_subscriber(email, display_name)
        except InvalidEmailAddressError as e:
            print('Invalid email address: {}'.format(e), file=results)
            return ContinueProcessing.yes
        try:
            ISubscriptionManager(mlist).register(subscriber,
                                                 delivery_mode=delivery_mode)
        except (AlreadySubscribedError, InvalidEmailAddressError,
                MembershipIsBannedError) as e:
            print(str(e), file=results)
            results.send_response = True
        except SubscriptionPendingError:
            # SubscriptionPendingError doesn't return an error message.
            listname = mlist.fqdn_listname                        # noqa: F841
            print(_('${person} has a pending subscription for ${listname}'),
                  file=results)
            results.send_response = True
        else:
            print(_('Confirmation email sent to ${person}'), file=results)
        return ContinueProcessing.yes

    def _parse_arguments(self, arguments, results):
        """Parse command arguments.

        :param arguments: The sequences of arguments as given to the
            `process()` method.
        :param results: The results object.
        :return: A tuple (delivery mode, address) or
            (ContinueProcessing.no, None) on error.
        """
        mode = DeliveryMode.regular
        address = None
        for argument in arguments:
            parts = argument.split('=', 1)
            if len(parts) != 2 or parts[0] not in ('digest', 'address'):
                print(self.name, _('bad argument: ${argument}'),
                      file=results)
                return (ContinueProcessing.no, None)
            if parts[0] == 'digest':
                mode = {
                    'no': DeliveryMode.regular,
                    'plain': DeliveryMode.plaintext_digests,
                    'mime': DeliveryMode.mime_digests,
                    }.get(parts[1])
                if mode is None:
                    print(self.name, _('bad argument: ${argument}'),
                          file=results)
                    return (ContinueProcessing.no, None)
            if parts[0] == 'address':
                address = parts[1]
        return (mode, address)


@public
class Subscribe(Join):
    """The email 'subscribe' command (an alias for 'join')."""

    name = 'subscribe'
    description = _("An alias for 'join'.")
    short_description = description


@public
@implementer(IEmailCommand)
class Leave:
    """The email 'leave' command."""

    name = 'leave'
    argument_description = ''
    description = _("""Leave this mailing list.

You may be asked to confirm your request.""")
    short_description = _('Leave this mailing list.')

    def process(self, mlist, msg, msgdata, arguments, results):
        """See `IEmailCommand`."""
        email = msg.sender
        if not email:
            print(
                _('${self.name}: No valid email address found to unsubscribe'),
                file=results)
            return ContinueProcessing.no
        user_manager = getUtility(IUserManager)
        user = user_manager.get_user(email)
        if user is None:
            print(_('No registered user for email address: ${email}'),
                  file=results)
            results.send_response = True
            return ContinueProcessing.no
        # The address that the -leave command was sent from, must be verified.
        # Otherwise you could link a bogus address to anyone's account, and
        # then send a leave command from that address.
        if user_manager.get_address(email).verified_on is None:
            print(_('Invalid or unverified email address: ${email}'),
                  file=results)
            results.send_response = True
            return ContinueProcessing.no
        already_left = msgdata.setdefault('leaves', set())
        for user_address in user.addresses:
            # Only recognize verified addresses.
            if user_address.verified_on is None:
                continue
            member = mlist.members.get_member(user_address.email)
            if member is not None:
                break
        else:
            # There are two possible situations.  Either none of the user's
            # addresses are subscribed to this mailing list, or this command
            # email *already* unsubscribed the user from the mailing list.
            # E.g. if a message was sent to the -leave address and it
            # contained the 'leave' command.  Don't send a bogus response in
            # this case, just ignore subsequent leaves of the same address.
            if email not in already_left:
                print(_('${self.name}: ${email} is not a member of '
                        '${mlist.fqdn_listname}'), file=results)
                return ContinueProcessing.no
        if email in already_left:
            return ContinueProcessing.yes
        # Ignore any subsequent 'leave' commands.
        already_left.add(email)
        manager = ISubscriptionManager(mlist)
        token, token_owner, member = manager.unregister(user_address)
        person = formataddr((user.display_name, email))   # noqa: F841
        if member is None:
            print(_('${person} left ${mlist.fqdn_listname}'), file=results)
        else:
            print(_('Confirmation email sent to ${person} to leave'
                    ' ${mlist.fqdn_listname}'), file=results)
        return ContinueProcessing.yes


@public
class Unsubscribe(Leave):
    """The email 'unsubscribe' command (an alias for 'leave')."""

    name = 'unsubscribe'
    description = _("An alias for 'leave'.")
    short_description = description



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test pendings."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.interfaces.pending import IPendable, IPendings
from mailman.interfaces.subscriptions import TokenOwner
from mailman.interfaces.workflow import IWorkflowStateManager
from mailman.model.pending import PendedKeyValue
from mailman.testing.layers import ConfigLayer
from zope.component import getUtility
from zope.interface import implementer


@implementer(IPendable)
class SimplePendable(dict):
    PEND_TYPE = 'simple'


class TestPendings(unittest.TestCase):
    """Test pendings."""

    layer = ConfigLayer

    def test_delete_key_values(self):
        # Deleting a pending should delete its key-values.
        pendingdb = getUtility(IPendings)
        subscription = SimplePendable(
            type='subscription',
            address='aperson@example.com',
            display_name='Anne Person',
            language='en',
            password='xyz')
        token = pendingdb.add(subscription)
        self.assertEqual(pendingdb.count(), 1)
        pendingdb.confirm(token)
        self.assertEqual(pendingdb.count(), 0)
        self.assertEqual(config.db.store.query(PendedKeyValue).count(), 0)

    def test_delete_workflow(self):
        # Deleting a pending should delete any associated workflow state.
        pendingdb = getUtility(IPendings)
        wsmanager = getUtility(IWorkflowStateManager)
        subscription = SimplePendable(
            type='subscription',
            address='aperson@example.com',
            display_name='Anne Person',
            language='en',
            password='xyz')
        token = pendingdb.add(subscription)
        wsmanager.save(token, step='step1', data='data')
        self.assertEqual(wsmanager.count, 1)
        pendingdb.confirm(token)
        self.assertEqual(wsmanager.count, 0)

    def test_find(self):
        # Test getting pendables for a mailing-list.
        mlist = create_list('list1@example.com')
        pendingdb = getUtility(IPendings)
        subscription_1 = SimplePendable(
            type='subscription',
            list_id='list1.example.com',
            token_owner='subscriber')
        subscription_2 = SimplePendable(
            type='subscription',
            list_id='list2.example.com')
        subscription_3 = SimplePendable(
            type='hold request',
            list_id='list1.example.com')
        subscription_4 = SimplePendable(
            type='hold request',
            list_id='list2.example.com')
        subscription_5 = SimplePendable(
            type='subscription',
            list_id='list2.example.com',
            token_owner='moderator')
        token_1 = pendingdb.add(subscription_1)
        pendingdb.add(subscription_2)
        token_3 = pendingdb.add(subscription_3)
        token_4 = pendingdb.add(subscription_4)
        token_5 = pendingdb.add(subscription_5)
        self.assertEqual(pendingdb.count(), 5)
        # Find the pending subscription in list1.
        pendings = list(pendingdb.find(mlist=mlist, pend_type='subscription'))
        self.assertEqual(len(pendings), 1, pendings)
        self.assertEqual(pendings[0][0], token_1)
        self.assertEqual(pendings[0][1]['list_id'], 'list1.example.com')
        self.assertEqual(pendings[0][1]['token_owner'], 'subscriber')
        # Find the pending subscription using the token_owner.
        pendings = list(pendingdb.find(token_owner=TokenOwner.moderator))
        self.assertEqual(len(pendings), 1)
        self.assertEqual(pendings[0][0], token_5)
        self.assertEqual(pendings[0][1]['list_id'], 'list2.example.com')
        self.assertEqual(pendings[0][1]['token_owner'], 'moderator')
        # Find all pending hold requests.
        pendings = list(pendingdb.find(pend_type='hold request'))
        self.assertEqual(len(pendings), 2)
        self.assertSetEqual(
            set((p[0], p[1]['list_id']) for p in pendings),
            {(token_3, 'list1.example.com'), (token_4, 'list2.example.com')}
            )
        # Find all pendings for list1.
        pendings = list(pendingdb.find(mlist=mlist))
        self.assertEqual(len(pendings), 2)
        self.assertSetEqual(
            set((p[0], p[1]['list_id'], p[1]['type']) for p in pendings),
            {(token_1, 'list1.example.com', 'subscription'),
             (token_3, 'list1.example.com', 'hold request')}
            )

        # Run count queries.
        self.assertEqual(
            pendingdb.count(mlist=mlist, pend_type='subscription'), 1)
        self.assertEqual(
            pendingdb.count(token_owner=TokenOwner.moderator), 1)
        self.assertEqual(
            pendingdb.count(token_owner=TokenOwner.subscriber), 1)
        self.assertEqual(
            pendingdb.count(mlist=mlist, token_owner=TokenOwner.subscriber), 1)



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interface for a mailing list."""

from enum import Enum
from mailman.interfaces.address import InvalidEmailAddressError
from mailman.interfaces.member import MemberRole
from public import public
from zope.interface import Attribute, Interface


@public
class InvalidListNameError(InvalidEmailAddressError):
    """List name is invalid."""

    def __init__(self, listname):
        super().__init__('{}@'.format(listname))
        self.listname = listname


@public
class DMARCMitigateAction(Enum):
    # Mitigations to apply to messages From: domains publishing an applicable
    # DMARC policy, or unconditionally depending on settings.
    #
    # No DMARC mitigations.
    no_mitigation = 0
    # Messages From: domains with DMARC policy will have From: replaced by the
    # list posting address and the original From: added to Reply-To: or Cc:.
    munge_from = 1
    # Messages From: domains with DMARC policy will be wrapped in an outer
    # message From: the list posting address.
    wrap_message = 2
    # Messages From: domains with DMARC policy will be rejected.
    reject = 3
    # Messages From: domains with DMARC policy will be discarded.
    discard = 4


@public
class Personalization(Enum):
    none = 0
    # Everyone gets a unique copy of the message, and there are a few more
    # substitution variables, but no headers are modified.
    individual = 1
    # All of the 'individual' personalization plus recipient header
    # modification.
    full = 2


@public
class ReplyToMunging(Enum):
    # The Reply-To header is passed through untouched
    no_munging = 0
    # The mailing list's posting address is appended to the Reply-To header
    point_to_list = 1
    # An explicit Reply-To header is added
    explicit_header = 2
    # An explicit Reply-To header is added and
    # the list address is removed from CC
    explicit_header_only = 3


@public
class SubscriptionPolicy(Enum):
    """All subscription/unsubscription policies for a mailing list."""
    # Neither confirmation, nor moderator approval is required.
    open = 0
    # The user must confirm the subscription.
    confirm = 1
    # The moderator must approve the subscription.
    moderate = 2
    # The user must first confirm their subscription, and then if that is
    # successful, the moderator must also approve it.
    confirm_then_moderate = 3


@public
class ArchiveRenderingMode(Enum):
    """Email rendering mode in Archiver."""
    # Default text.
    text = 1
    # Render emails as markdown text.
    markdown = 2


@public
class IMailingList(Interface):
    """A mailing list."""

    # List identity

    created_at = Attribute(
        """The date and time that the mailing list was created.""")

    list_name = Attribute("""\
        The read-only short name of the mailing list.  Note that where a
        Mailman installation supports multiple domains, this short name may
        not be unique.  Use the fqdn_listname attribute for a guaranteed
        unique id for the mailing list.  This short name is always the local
        part of the posting email address.  For example, if messages are
        posted to mylist@example.com, then the list_name is 'mylist'.
        """)

    mail_host = Attribute("""\
        The read-only domain name 'hosting' this mailing list.  This is always
        the domain name part of the posting email address, and it may bear no
        relationship to the web url used to access this mailing list.  For
        example, if messages are posted to mylist@example.com, then the
        mail_host is 'example.com'.
        """)

    list_id = Attribute("""\
        The identity of the mailing list.  This value will never change.  It
        is defined in RFC 2369.
        """)

    fqdn_listname = Attribute("""\
        The read-only fully qualified name of the mailing list.  This is the
        guaranteed unique id for the mailing list, and it is always the
        address to which messages are posted, e.g. mylist@example.com.  It is
        always comprised of the list_name + '@' + mail_host.
        """)

    domain = Attribute(
        """The `IDomain` that this mailing list is defined in.""")

    display_name = Attribute("""\
        The short human-readable descriptive name for the mailing list.  This
        is used in locations such as the message footers and as the default
        value for the Subject prefix.
        """)

    description = Attribute("""\
        A terse phrase identifying this mailing list.

        This description is used when the mailing list is listed with other
        mailing lists, or in headers, and so forth.  It should be as succinct
        as you can get it, while still identifying what the list is.""")

    info = Attribute("""\
        A longer description of this mailing list.  This can be any arbitrary
        text, up to a database-specific maximum length.
        """)

    preferred_language = Attribute("""\
        The default language for communications on this mailing list.

        When the list sends out notifications, it will be in this language,
        unless the recipient is a known user and that user has a preferred
        language.
        """)

    subject_prefix = Attribute("""\
        The text to insert at the front of the Subject field.

        When messages posted to this mailing list are sent to the list
        subscribers, the Subject header may be rewritten to include an
        identifying prefix.  Typically this prefix will appear in square
        brackets and the default value inside the brackets is taken as the
        list's display name.  However, any value can be used, including the
        empty string to prevent Subject header rewriting.
        """)

    allow_list_posts = Attribute(
        """Flag specifying posts to the list are generally allowed.

        This controls the value of the RFC 2369 List-Post header.  This is
        usually set to True, except for announce-only lists.  When False, the
        List-Post is set to NO as per the RFC.
        """)

    include_rfc2369_headers = Attribute(
        """Flag specifying whether to include any RFC 2369 header, including
        the RFC 2919 List-ID header.""")

    anonymous_list = Attribute(
        """Flag controlling whether messages to this list are anonymized.

        Anonymizing messages is not perfect, however setting this flag removes
        the sender of the message (in the From, Sender, and Reply-To fields)
        and replaces these with the list's posting address.
        """)

    advertised = Attribute(
        """Advertise this mailing list when people ask for an overview of the
        available mailing lists.""")

    # Contact addresses

    posting_address = Attribute(
        """The address to which messages are posted for copying to the full
        list membership, where 'full' of course means those members for which
        delivery is currently enabled.
        """)

    no_reply_address = Attribute(
        """The address to which all messages will be immediately discarded,
        without prejudice or record.  This address is specific to the domain,
        even though it's available on the IMailingListAddresses interface.
        Generally, humans should never respond directly to this address.
        """)

    owner_address = Attribute(
        """The address which reaches the owners and moderators of the mailing
        list.  There is no address which reaches just the owners or just the
        moderators of a mailing list.
        """)

    request_address = Attribute(
        """The address which reaches the email robot for this mailing list.
        This robot can process various email commands such as changing
        delivery options, getting information or help about the mailing list,
        or processing subscrptions and unsubscriptions (although for the
        latter two, it's better to use the join_address and leave_address.
        """)

    bounces_address = Attribute(
        """The address which reaches the automated bounce processor for this
        mailing list.  Generally, humans should never respond directly to this
        address.
        """)

    join_address = Attribute(
        """The address to which subscription requests should be sent.""")

    leave_address = Attribute(
        """The address to which unsubscription requests should be sent.""")

    subscribe_address = Attribute(
        """Deprecated address to which subscription requests may be sent.
        This address is provided for backward compatibility only.  See
        `join_address` for the preferred alias.
        """)

    unsubscribe_address = Attribute(
        """Deprecated address to which unsubscription requests may be sent.
        This address is provided for backward compatibility only.  See
        `leave_address` for the preferred alias.
        """)

    def confirm_address(cookie=''):
        """The address used for various forms of email confirmation."""

    # DMARC attributes.

    dmarc_mitigate_action = Attribute(
        """The mitigation to apply to messages from a DMARC matching domain.

        This is a  DMARCMitigateAction to be applied to messages From: a domain
        publishing DMARC p=reject or quarantine, and possibly unconditionally
        depending on the setting of dmarc_mitigate_unconditionally.
        """)

    dmarc_mitigate_unconditionally = Attribute(
        """Should DMARC mitigations apply unconditionally?

        A flag indicating whether to apply dmarc_mitigate_action to all
        messages, but only if dmarc_mitigate_action is other than reject or
        discard.
        """)

    dmarc_moderation_notice = Attribute(
        """Text to include in any DMARC rejected message.

        Rejection notices are sent when DMARCMitigateAction of reject applies.
        """)

    dmarc_wrapped_message_text = Attribute(
        """Additional MIME text to include in DMARC wrapped messages.

        This text is added as a separate text/plain MIME part preceding the
        original message part in the wrapped message when DMARCMitigateAction
        of wrap_message applies.
        """)

    # Rosters and subscriptions.

    owners = Attribute(
        """The IUser owners of this mailing list.

        This does not include the IUsers who are moderators but not owners of
        the mailing list.""")

    moderators = Attribute(
        """The IUser moderators of this mailing list.

        This does not include the IUsers who are owners but not moderators of
        the mailing list.""")

    administrators = Attribute(
        """The IUser administrators of this mailing list.

        This includes the IUsers who are both owners and moderators of the
        mailing list.""")

    nonmembers = Attribute(
        """A roster of all the nonmembers of the mailing list.""")

    members = Attribute(
        """A roster of all the members of the mailing list, regardless of
        whether they are to receive regular messages or digests, or whether
        they have their delivery disabled or not.""")

    regular_members = Attribute(
        """A roster of all the IMembers who are to receive regular postings
        (i.e. non-digests) from the mailing list, regardless of whether they
        have their delivery disabled or not.""")

    digest_members = Attribute(
        """A roster of all the IMembers who are to receive digests of postings
        to this mailing list, regardless of whether they have their deliver
        disabled or not, or of the type of digest they are to receive.""")

    subscription_policy = Attribute(
        """The policy for subscribing new members to the list.""")

    unsubscription_policy = Attribute(
        """The policy for unsubscribing members from the list.""")

    subscribers = Attribute(
        """An iterator over all IMembers subscribed to this list, with any
        role.
        """)

    def get_roster(role):
        """Return the appropriate roster for the given role.

        :param role: The requested roster's role.
        :type role: MemberRole
        :return: The requested roster.
        :rtype: Roster
        """

    def is_subscribed(subscriber, role=MemberRole.member):
        """Is the given address or user subscribed to the mailing list?

        :param subscriber: The address or user to check.
        :type subscriber: `IUser` or `IAddress`
        :param role: The role being checked (e.g. a member, owner, or
            moderator of a mailing list).
        :type role: `MemberRole`
        :return: A flag indicating whether the subscriber is already
            subscribed to the mailing list or not.
        """

    def subscribe(subscriber, role=MemberRole.member):
        """Subscribe the given address or user to the mailing list.

        :param subscriber: The address or user to subscribe to the mailing
            list.  The user's preferred address receives deliveries, if she
            has one, otherwise no address for the user appears in the rosters.
        :type subscriber: `IUser` or `IAddress`
        :param role: The role being subscribed to (e.g. a member, owner, or
            moderator of a mailing list).
        :type role: `MemberRole`
        :return: The member object representing the subscription.
        :rtype: `IMember`
        :raises AlreadySubscribedError: If the address or user is already
            subscribed to the mailing list with the given role.  Note however
            that it is possible to subscribe an address to a mailing list with
            a particular role, and also subscribe a user with a matching
            preferred address that is explicitly subscribed with the same role.
        :raises InvalidEmailAddressError: If the address being subscribed is
            the list's posting address.
        """

    # Delivery.

    archive_policy = Attribute(
        """The policy for archiving messages to this mailing list.

        The value is an `ArchivePolicy` enum.  Use this to archive the mailing
        list publicly, privately, or not at all.
        """)

    last_post_at = Attribute(
        """The date and time a message was last posted to the mailing list.""")

    post_id = Attribute(
        """A monotonically increasing integer sequentially assigned to each
        list posting.""")

    personalize = Attribute(
        """The type of personalization that is applied to postings.""")

    reply_goes_to_list = Attribute(
        """Reply-To: header munging policy.""")

    member_roster_visibility = Attribute(
        """The policy for who can view the member roster of this mailing list.

        The value is a `RosterVisibility` enum.  Use this to change who can
        view the member list. Options are public, members, or moderators.""")

    # Digests.

    digests_enabled = Attribute(
        """Whether or not digests are enabled for this mailing list.""")

    digest_size_threshold = Attribute(
        """The maximum (approximate) size in kilobytes of the digest currently
        being collected.""")

    digest_send_periodic = Attribute(
        """Should a digest be sent by the `mailman send_digest` command even
        when the size threshold hasn't yet been met?""")

    digest_volume_frequency = Attribute(
        """How often should a new digest volume be started?""")

    digest_last_sent_at = Attribute(
        """The date and time a digest of this mailing list was last sent.""")

    volume = Attribute(
        """A monotonically increasing integer sequentially assigned to each
        new digest volume.  The volume number may be bumped either
        automatically (i.e. on a defined schedule) or manually.  When the
        volume number is bumped, the digest number is always reset to 1.""")

    next_digest_number = Attribute(
        """A sequence number for a specific digest in a given volume.  When
        the digest volume number is bumped, the digest number is reset to
        1.""")

    def send_one_last_digest_to(address, delivery_mode):
        """Make sure to send one last digest to an address.

        This is used when a person transitions from digest delivery to regular
        delivery and wants to make sure they don't miss anything.  By
        indicating that they'd like to receive one last digest, they will
        ensure continuity in receiving mailing lists posts.

        :param address: The address of the person receiving one last digest.
        :type address: `IAddress`
        :param delivery_mode: The type of digest to receive.
        :type delivery_mode: `DeliveryMode`
        """

    last_digest_recipients = Attribute(
        """An iterator over the addresses that should receive one last digest.

        Items are 2-tuples of (`IAddress`, `DeliveryMode`).  The one last
        digest recipients are cleared.
        """)

    # Autoresponses.

    autoresponse_grace_period = Attribute(
        """Time period (in days) between automatic responses.

        When this mailing list is set to send an auto-response for messages
        sent to mailing list posts, the mailing list owners, or the `-request`
        address, such reponses will not be sent to the same user more than
        once during the grace period.  Set to zero (or a negative value) for
        no grace period (i.e. auto-respond to every message).
        """)

    autorespond_owner = Attribute(
        """How should the mailing list automatically respond to messages sent
        to the -owner or -moderator address?

        Options are:
        * No response sent
        * Send a response and discard the original messge
        * Send a response and continue processing the original message
        """)

    autoresponse_owner_text = Attribute(
        """The text sent in an autoresponse to the owner or moderator.""")

    autorespond_postings = Attribute(
        """How should the mailing list automatically respond to messages sent
        to the list's posting address?

        Options are:
        * No response sent
        * Send a response and discard the original messge
        * Send a response and continue processing the original message
        """)

    autoresponse_postings_text = Attribute(
        """The text sent in an autoresponse to the list's posting address.""")

    autorespond_requests = Attribute(
        """How should the mailing list automatically respond to messages sent
        to the list's `-request` address?

        Options are:
        * No response sent
        * Send a response and discard the original messge
        * Send a response and continue processing the original message
        """)

    autoresponse_request_text = Attribute(
        """The text sent in an autoresponse to the list's `-request`
        address.""")

    # Processing.

    posting_chain = Attribute(
        """This mailing list's posting moderation chain.

        When messages are posted to a mailing list, it first goes through a
        moderation chain to determine whether the message will be accepted.
        This attribute names a chain for postings, which must exist.
        """)

    posting_pipeline = Attribute(
        """This mailing list's posting pipeline.

        Every mailing list has a processing pipeline that messages flow
        through once they've been accepted for posting to the mailing list.
        This attribute names a pipeline for postings, which must exist.
        """)

    owner_chain = Attribute(
        """This mailing list's owner moderation chain.

        When messages are posted to the owners of a mailing list, it first
        goes through a moderation chain to determine whether the message will
        be accepted.  This attribute names a chain for postings, which must
        exist.
        """)

    owner_pipeline = Attribute(
        """This mailing list's owner posting pipeline.

        Every mailing list has a processing pipeline that messages flow
        through once they've been accepted for posting to the owners of a
        mailing list.  This attribute names a pipeline for postings, which
        must exist.
        """)

    data_path = Attribute(
        """The file system path to list-specific data.

        An example of list-specific data is the temporary digest mbox file
        that gets created to accumlate messages for the digest.
        """)

    require_explicit_destination = Attribute(
        """Flag controlling requiring explisit destination.

        If require_explicit_destination is True, a post must have the list's
        posting address or an acceptable alias in To: or Cc: or be held for
        moderator approval.
        """)

    acceptable_aliases = Attribute(
        """List of addresses and regexps acceptable as explicit destination.

        This is a list of addresses and regexps matching addresses that are
        acceptable in To: or Cc: when require_explicit_destination is True.
        """)

    administrivia = Attribute(
        """Flag controlling `administrivia` checks.

        Administrivia tests check whether postings to the mailing list are
        really meant for the -request address.  Examples include messages with
        `help`, `subscribe`, `unsubscribe`, and other commands.  When such
        messages are incorrectly posted to the general mailing list, they are
        just noise, and when this flag is set will be intercepted and in
        general held for moderator approval.
        """)

    filter_content = Attribute(
        """Flag specifying whether to filter a message's content.

        Filtering is performed on MIME type and file name extension.
        """)

    filter_action = Attribute(
        """Action to take when the top-level content-type is filtered.

        The value is a `FilterAction` enum.
        """)

    convert_html_to_plaintext = Attribute(
        """Flag specifying whether text/html parts should be converted.

        When True, after filtering, if there are any text/html parts left in
        the original message, they are converted to text/plain.
        """)

    collapse_alternatives = Attribute(
        """Flag specifying whether multipart/alternatives should be collapsed.

        After all MIME content filtering is complete, collapsing alternatives
        replaces the outer multipart/alternative parts with the first
        subpart.
        """)

    filter_types = Attribute(
        """Sequence of MIME types that should be filtered out.

        These can be either main types or main/sub types.  Set this attribute
        to a sequence to change it, or to None to empty it.
        """)

    pass_types = Attribute(
        """Sequence of MIME types to explicitly pass.

        These can be either main types or main/sub types.  Set this attribute
        to a sequence to change it, or to None to empty it.  Pass types are
        consulted after filter types, and only if `pass_types` is non-empty.
        """)

    filter_extensions = Attribute(
        """Sequence of file extensions that should be filtered out.

        Set this attribute to a sequence to change it, or to None to empty it.
        """)

    pass_extensions = Attribute(
        """Sequence of file extensions to explicitly pass.

        Set this attribute to a sequence to change it, or to None to empty it.
        Pass extensions are consulted after filter extensions, and only if
        `pass_extensions` is non-empty.
        """)

    # Moderation.

    default_member_action = Attribute(
        """The default action to take for postings from members.

        When an address is subscribed to the mailing list, this attribute sets
        the initial moderation action (as an `Action`).  When the action is
        `Action.defer` (the default), then normal posting decisions are made.
        When the action is `Action.accept`, the postings are accepted without
        any other checks.
        """)

    default_nonmember_action = Attribute(
        """The default action to take for postings from nonmembers.

        When a nonmember address posts to the mailing list, this attribute
        sets the initial moderation action (as an `Action`).  When the action
        is `Action.defer` (the default), then normal posting decisions are
        made.  When the action is `Action.accept`, the postings are accepted
        without any other checks.
        """)

    # Bounces.
    bounce_info_stale_after = Attribute(
        """Number of days after which bounce info is considered stale.

        The number of days after which a member's bounce information is
        considered stale. If no new bounces have been received in the interim,
        the bounce score is reset to zero. This value must be an integer.
        """)

    bounce_notify_owner_on_bounce_increment = Attribute(
        """This option controls whether or not the list owner
        is notified when a member's bounce score is incremented.
        """)

    bounce_notify_owner_on_disable = Attribute(
        """This option controls whether or not the list owner
        is notified when a member's subscription is automatically
        disabled due to their bounce threshold being reached.
        """)

    bounce_notify_owner_on_removal = Attribute(
        """This option controls whether or not the list owner
        is notified when a member is removed from the list after
        their disabled notifications have been exhausted.
        """)

    bounce_score_threshold = Attribute(
        """Bounce score threshold for a mailing-list.

        Threshold value for a Member's bounce_score after which we will either
        disable membership or unsubscribe the member from the mailing list.
        """)

    bounce_you_are_disabled_warnings = Attribute(
        """The number of notices a disabled member will
        receive before their address is removed from the
        mailing list's roster. Set this to 0 to immediately
        remove an address from the list once their bounce
        score exceeds the threshold. This value must be an integer.
        """)

    bounce_you_are_disabled_warnings_interval = Attribute(
        """The number of days between each disabled notification.""")

    forward_unrecognized_bounces_to = Attribute(
        """What to do when a bounce contains no recognizable addresses.

        This is an enumeration which specifies what to do with such bounce
        messages.  They can be discarded, forward to the list owner, or
        forwarded to the site owner.
        """)

    process_bounces = Attribute(
        """Whether or not the mailing list processes bounces.""")

    # Notifications.

    admin_immed_notify = Attribute(
        """Flag controlling immediate notification of requests.

        List moderators normally get daily notices about pending
        administrative requests.  This flag controls whether moderators also
        receive immediate notification of such pending requests.
        """)

    admin_notify_mchanges = Attribute(
        """Flag controlling notification of joins and leaves.

        List moderators can receive notifications for every member that joins
        or leaves their mailing lists.  This flag controls those
        notifications.
        """)

    send_welcome_message = Attribute(
        """Flag indicating whether a welcome message should be sent.""")

    send_goodbye_message = Attribute(
        """Flag indicating whether a goodbye message should be sent.""")

    # Usenet gateway.

    gateway_to_mail = Attribute(
        """Flag indicating that posts to the linked newsgroup should be gated
        to the list.""")

    gateway_to_news = Attribute(
        """Flag indicating that posts to the list should be gated to the
        linked newsgroup.""")

    linked_newsgroup = Attribute(
        """The name of the linked newsgroup.""")

    newsgroup_moderation = Attribute(
        """The moderation policy for the linked newsgroup, if there is one.""")

    nntp_prefix_subject_too = Attribute(
        """Flag indicating whether the list's subject_prefix should be included
        in posts gated to usenet.""")

    usenet_watermark = Attribute(
        """The NNTP server's message number of the last post gated to the
        list.""")

    # Hyperkitty specific settings.

    archive_rendering_mode = Attribute(
        """How are Emails rendered in archivers?  This is only a suggested mode
        for Archivers and is not related to actual contents of Emails.

        Primary use of this setting is going to be Hyperkitty, which can render
        emails as markdown text.

        Values are `ArchiveRenderingMode` Enum.
        """)


@public
class IAcceptableAlias(Interface):
    """An acceptable alias for implicit destinations."""

    mailing_list = Attribute('The associated mailing list.')

    address = Attribute('The address or pattern to match against recipients.')


@public
class IAcceptableAliasSet(Interface):
    """The set of acceptable aliases for a mailing list."""

    def clear():
        """Clear the set of acceptable posting aliases."""

    def add(alias):
        """Add the given address as an acceptable aliases for posting.

        :param alias: The email address to accept as a recipient for implicit
            destination posting purposes.  The alias is coerced to lower
            case.  If `alias` begins with a '^' character, it is interpreted
            as a regular expression, otherwise it must be an email address.
        :type alias: string
        :raises ValueError: when the alias neither starts with '^' nor has an
            '@' sign in it.
        """

    def remove(alias):
        """Remove the given address as an acceptable aliases for posting.

        :param alias: The email address to no longer accept as a recipient for
            implicit destination posting purposes.
        :type alias: string
        """

    aliases = Attribute(
        """An iterator over all the acceptable aliases.""")


@public
class IListArchiver(Interface):
    """An archiver for a mailing list.

    The named archiver must be enabled site-wide in order for a mailing list
    to be able to enable it.
    """

    mailing_list = Attribute('The associated mailing list.')

    name = Attribute('The name of the archiver.')

    is_enabled = Attribute('Is this archiver enabled for this mailing list?')

    system_archiver = Attribute(
        'The associated system-wide IArchiver instance.')


@public
class IListArchiverSet(Interface):
    """The set of archivers (enabled or disabled) for a mailing list."""

    archivers = Attribute(
        """An iterator over all the archivers for this mailing list.""")

    def get(archiver_name):
        """Return the `IListArchiver` with the given name, if it exists.

        :param archiver_name: The name of the archiver.
        :type archiver_name: unicode.
        :return: the matching `IListArchiver` or None if the named archiver
            does not exist.
        """


@public
class IHeaderMatch(Interface):
    """A mailing list-specific message header matching rule."""

    mailing_list = Attribute(
        """The mailing list for the header match.""")

    header = Attribute(
        """The email header that will be checked.""")

    pattern = Attribute(
        """The regular expression to match.""")

    position = Attribute(
        """The ordinal position of this header match.

        Set this to change the position of this header match.
        """)

    chain = Attribute(
        """The chain to jump to on a match.

        If it is None, the `[antispam]jump_chain` action in the configuration
        file is used.
        """)

    tag = Attribute(
        """An arbitrary value to identify a set of IHeaderMatches.

        This tag can be used to filter a set of IHeaderMatches, so that they
        can set and removed together, without having to query the whole list
        and iterate over them.
        """)


@public
class IHeaderMatchList(Interface):
    """The list of header matching rules for a mailing list."""

    def clear():
        """Clear the list of header matching rules."""

    def append(header, pattern, chain=None, tag=None):
        """Append the given rule to this mailing list's header match list.

        :param header: The email header to filter on.  It will be converted to
            lower case for consistency.
        :type header: string
        :param pattern: The regular expression to use.
        :type pattern: string
        :param chain: The chain to jump to, or None to use the site-wide
            antispam jump chain via the configuration.  Defaults to None.
        :type chain: string or None
        :param tag: An arbitrary value to identify a set of IHeaderMatches.
        :type tag: string or None
        :raises ValueError: if the header/pattern pair already exists for this
            mailing list.
        """

    def insert(index, header, pattern, chain=None, tag=None):
        """Insert a header match rule.

        Inserts the given rule at the given index position in this
        mailing list's header match list.

        :param index: The index to insert the rule at.
        :type index: integer
        :param header: The email header to filter on.  It will be converted to
            lower case for consistency.
        :type header: string
        :param pattern: The regular expression to use.
        :type pattern: string
        :param chain: The chain to jump to, or None to use the site-wide
            antispam jump chain via the configuration.  Defaults to None.
        :type chain: string or None
        :param tag: An arbitrary value to identify a set of IHeaderMatches.
        :type tag: string or None
        :raises ValueError: if the header/pattern pair already exists for this
            mailing list.
        """

    def remove(header, pattern):
        """Remove the given rule from this mailing list's header match list.

        :param header: The email header part of the rule to be removed.
        :type header: string
        :param pattern: The regular expression part of the rule to be removed.
        :type pattern: string
        :raises ValueError: if the header does not exist in the list of
            header matches.
        """

    def __getitem__(index):
        """Return the header match at the given index for this mailing list.

        :param index: The index of the header match to return.
        :type index: integer
        :return: The header match at this index.
        :rtype: `IHeaderMatch`
        :raises IndexError: if there is no header match at this index for
            this mailing list.
        """

    def __delitem__(index):
        """Remove the rule at the given index.

        :param index: The index of the header match to remove.
        :type index: integer
        :raises IndexError: if there is no header match at this index for
            this mailing list.
        """

    def __len__():
        """Return the number of header matches for this mailing list.

        :rtype: integer
        """

    def __iter__():
        """An iterator over all the IHeaderMatches defined in this list.

        :return: iterator over `IHeaderMatch`.
        """



# Copyright (C) 2000-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Archive runner."""

import copy
import logging

from datetime import datetime
from email.utils import mktime_tz, parsedate_tz
from lazr.config import as_timedelta
from mailman.config import config
from mailman.core.runner import Runner
from mailman.interfaces.archiver import ClobberDate
from mailman.interfaces.mailinglist import IListArchiverSet
from mailman.utilities.datetime import now, RFC822_DATE_FMT
from public import public


log = logging.getLogger('mailman.archiver')


def _should_clobber(msg, msgdata, archiver):
    """Should the Date header in the original message get clobbered?"""
    # Calculate the Date header of the message as a datetime.  What if there
    # are multiple Date headers, even in violation of the RFC?  For now, take
    # the first one.  If there are no Date headers, then definitely clobber.
    original_date = msg.get('date')
    if original_date is None:
        return True
    section = getattr(config.archiver, archiver, None)
    if section is None:
        log.error('No archiver config section found: {}'.format(archiver))
        return False
    try:
        clobber = ClobberDate[section.clobber_date]
    except ValueError:
        log.error('Invalid clobber_date for "{}": {}'.format(
            archiver, section.clobber_date))
        return False
    if clobber is ClobberDate.always:
        return True
    elif clobber is ClobberDate.never:
        return False
    # Maybe we'll clobber the date.  Let's see if it's farther off from now
    # than the skew period.
    skew = as_timedelta(section.clobber_skew)
    try:
        time_tuple = parsedate_tz(original_date)
    except (ValueError, OverflowError):
        # The likely cause of this is that the year in the Date: field is
        # horribly incorrect, e.g. (from SF bug # 571634):
        #
        # Date: Tue, 18 Jun 0102 05:12:09 +0500
        #
        # Obviously clobber such dates.
        return True
    if time_tuple is None:
        # There was some other bogosity in the Date header.
        return True
    claimed_date = datetime.fromtimestamp(mktime_tz(time_tuple))
    return (abs(now() - claimed_date) > skew)


@public
class ArchiveRunner(Runner):
    """The archive runner."""

    def _dispose(self, mlist, msg, msgdata):
        received_time = msgdata.get('received_time', now(strip_tzinfo=False))
        archiver_set = IListArchiverSet(mlist)
        for archiver in archiver_set.archivers:
            # The archiver is disabled if either the list-specific or
            # site-wide archiver is disabled.
            if not archiver.is_enabled:
                continue
            msg_copy = copy.deepcopy(msg)
            if _should_clobber(msg, msgdata, archiver.name):
                original_date = msg_copy['date']
                del msg_copy['date']
                del msg_copy['x-original-date']
                msg_copy['Date'] = received_time.strftime(RFC822_DATE_FMT)
                if original_date:
                    msg_copy['X-Original-Date'] = original_date
            # A problem in one archiver should not prevent other archivers
            # from running.
            try:
                archiver.system_archiver.archive_message(mlist, msg_copy)
            except Exception:
                log.exception('Exception in "{}" archiver'.format(
                    archiver.name))



# Copyright (C) 2008-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Prototypical permalinking archiver."""

import os
import logging

from contextlib import suppress
from datetime import timedelta
from flufl.lock import Lock, TimeOutError
from mailbox import Maildir
from mailman.config import config
from mailman.interfaces.archiver import IArchiver
from public import public
from zope.interface import implementer


log = logging.getLogger('mailman.error')


@public
@implementer(IArchiver)
class Prototype:
    """A prototype of a third party archiver.

    Mailman proposes a draft specification for interoperability between list
    servers and archivers: <http://wiki.list.org/display/DEV/Stable+URLs>.
    """

    name = 'prototype'
    is_enabled = False

    @staticmethod
    def list_url(mlist):
        """See `IArchiver`."""
        # This archiver is not web-accessible, therefore no URL is returned.
        return None

    @staticmethod
    def permalink(mlist, msg):
        """See `IArchiver`."""
        # This archiver is not web-accessible, therefore no URL is returned.
        return None

    @staticmethod
    def archive_message(mlist, message):
        """See `IArchiver`.

        This archiver saves messages into a maildir.
        """
        archive_dir = os.path.join(config.ARCHIVE_DIR, 'prototype')
        with suppress(FileExistsError):
            os.makedirs(archive_dir, 0o775)
        # Maildir will throw an error if the directories are partially created
        # (for instance the toplevel exists but cur, new, or tmp do not)
        # therefore we don't create the toplevel as we did above.
        list_dir = os.path.join(archive_dir, mlist.fqdn_listname)
        mailbox = Maildir(list_dir, create=True, factory=None)
        lock_file = os.path.join(
            config.LOCK_DIR, '{0}-maildir.lock'.format(mlist.fqdn_listname))
        # Lock the maildir as Maildir.add() is not threadsafe.  Don't use the
        # context manager because it's not an error if we can't acquire the
        # archiver lock.  We'll just log the problem and continue.
        #
        # XXX 2012-03-14 BAW: When we extend the chain/pipeline architecture
        # to other runners, e.g. the archive runner, it would be better to let
        # any TimeOutError propagate up.  That would cause the message to be
        # re-queued and tried again later, rather than being discarded as
        # happens now below.
        lock = Lock(lock_file)
        try:
            lock.lock(timeout=timedelta(seconds=1))
            # Add the message to the maildir.  The return value could be used
            # to construct the file path if necessary.  E.g.
            #
            # os.path.join(archive_dir, mlist.fqdn_listname, 'new',
            #              message_key)
            mailbox.add(message)
        except TimeOutError:
            # Log the error and go on.
            log.error('Unable to acquire prototype archiver lock for {0}, '
                      'discarding: {1}'.format(
                          mlist.fqdn_listname,
                          message.get('message-id', 'n/a')))
        finally:
            lock.unlock(unconditionally=True)
        return None



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test for unsubscription service."""

import unittest

from contextlib import suppress
from mailman.app.lifecycle import create_list
from mailman.app.subscriptions import UnSubscriptionWorkflow
from mailman.interfaces.mailinglist import SubscriptionPolicy
from mailman.interfaces.pending import IPendings
from mailman.interfaces.subscriptions import TokenOwner
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import (
    get_queue_messages,
    LogFileMark,
    set_preferred,
    subscribe,
)
from mailman.testing.layers import ConfigLayer
from mailman.utilities.datetime import now
from unittest.mock import patch
from zope.component import getUtility


class TestUnSubscriptionWorkflow(unittest.TestCase):
    layer = ConfigLayer
    maxDiff = None

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._mlist.admin_immed_notify = False
        self._mlist.unsubscription_policy = SubscriptionPolicy.open
        self._mlist.send_welcome_message = False
        self._anne = 'anne@example.com'
        self._user_manager = getUtility(IUserManager)
        self.anne = self._user_manager.create_user(self._anne, 'Ане')
        self.anne.addresses[0].verified_on = now()
        self.anne.preferred_address = self.anne.addresses[0]
        self._mlist.subscribe(self.anne)
        self._expected_pendings_count = 0

    def tearDown(self):
        # There usually should be no pending after all is said and done, but
        # some tests don't complete the workflow.
        self.assertEqual(getUtility(IPendings).count(),
                         self._expected_pendings_count)

    def test_start_state(self):
        # Test the workflow starts with no tokens or members.
        workflow = UnSubscriptionWorkflow(self._mlist)
        self.assertEqual(workflow.token_owner, TokenOwner.no_one)
        self.assertIsNone(workflow.token)
        self.assertIsNone(workflow.member)

    def test_pended_data(self):
        # Test there is a Pendable object associated with a held
        # unsubscription request and it has some valid data associated with
        # it.
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        with suppress(StopIteration):
            workflow.run_thru('send_confirmation')
        self.assertIsNotNone(workflow.token)
        pendable = getUtility(IPendings).confirm(workflow.token, expunge=False)
        self.assertEqual(pendable['list_id'], 'test.example.com')
        self.assertEqual(pendable['email'], 'anne@example.com')
        self.assertEqual(pendable['display_name'], 'Ане')
        self.assertEqual(pendable['when'], '2005-08-01T07:49:23')
        self.assertEqual(pendable['token_owner'], 'subscriber')
        # The token is still in the database.
        self._expected_pendings_count = 1

    def test_user_or_address_required(self):
        # The `subscriber` attribute must be a user or address that is provided
        # to the workflow.
        workflow = UnSubscriptionWorkflow(self._mlist)
        self.assertRaises(AssertionError, list, workflow)

    def test_user_is_subscribed_to_unsubscribe(self):
        # A user must be subscribed to a list when trying to unsubscribe.
        addr = self._user_manager.create_address('aperson@example.org')
        addr.verfied_on = now()
        workflow = UnSubscriptionWorkflow(self._mlist, addr)
        self.assertRaises(AssertionError,
                          workflow.run_thru, 'subscription_checks')

    def test_subscription_checks_for_user(self):
        # subscription_checks must pass for IUser subscribed as IAddress.
        member = subscribe(self._mlist, 'Bart')
        set_preferred(member.user)
        workflow = UnSubscriptionWorkflow(self._mlist, member.user)
        workflow.run_thru('subscription_checks')

    def test_subscription_checks_for_address(self):
        # subscription_checks must pass for IAddress subscribed as IUser.
        workflow = UnSubscriptionWorkflow(self._mlist,
                                          self.anne.preferred_address)
        workflow.run_thru('subscription_checks')

    def test_confirmation_checks_open_list(self):
        # An unsubscription from an open list does not need to be confirmed or
        # moderated.
        self._mlist.unsubscription_policy = SubscriptionPolicy.open
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        workflow.run_thru('confirmation_checks')
        with patch.object(workflow, '_step_do_unsubscription') as step:
            next(workflow)
        step.assert_called_once_with()

    def test_confirmation_checks_no_user_confirmation_needed(self):
        # An unsubscription from a list which does not need user confirmation
        # skips to the moderation checks.
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne,
                                          pre_confirmed=True)
        workflow.run_thru('confirmation_checks')
        with patch.object(workflow, '_step_moderation_checks') as step:
            next(workflow)
        step.assert_called_once_with()

    def test_confirmation_checks_confirm_pre_confirmed(self):
        # The unsubscription policy requires user-confirmation, but their
        # unsubscription is pre-confirmed. Since moderation is not reuqired,
        # the user will be immediately unsubscribed.
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        workflow = UnSubscriptionWorkflow(
            self._mlist, self.anne, pre_confirmed=True)
        workflow.run_thru('confirmation_checks')
        with patch.object(workflow, '_step_do_unsubscription') as step:
            next(workflow)
        step.assert_called_once_with()

    def test_confirmation_checks_confirm_then_moderate_pre_confirmed(self):
        # The unsubscription policy requires user confirmation, but their
        # unsubscription is pre-confirmed. Since moderation is required, that
        # check will be performed.
        self._mlist.unsubscription_policy = (
            SubscriptionPolicy.confirm_then_moderate)
        workflow = UnSubscriptionWorkflow(
            self._mlist, self.anne, pre_confirmed=True)
        workflow.run_thru('confirmation_checks')
        with patch.object(workflow, '_step_do_unsubscription') as step:
            next(workflow)
        step.assert_called_once_with()

    def test_send_confirmation_checks_confirm_list(self):
        # The unsubscription policy requires user confirmation and the
        # unsubscription is not pre-confirmed.
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        workflow.run_thru('confirmation_checks')
        with patch.object(workflow, '_step_send_confirmation') as step:
            next(workflow)
        step.assert_called_once_with()

    def test_moderation_checks_moderated_list(self):
        # The unsubscription policy requires moderation.
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        workflow.run_thru('confirmation_checks')
        with patch.object(workflow, '_step_moderation_checks') as step:
            next(workflow)
            step.assert_called_once_with()

    def test_moderation_checks_approval_required(self):
        # The moderator must approve the subscription request.
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        workflow.run_thru('moderation_checks')
        with patch.object(workflow, '_step_get_moderator_approval') as step:
            next(workflow)
        step.assert_called_once_with()

    def test_do_unsusbcription(self):
        # An open unsubscription policy means the user gets unsubscribed to
        # the mailing list without any further confirmations or approvals.
        self._mlist.unsubscription_policy = SubscriptionPolicy.open
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        list(workflow)
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNone(member)

    def test_do_unsubscription_pre_approved(self):
        # A moderation-requiring subscription policy plus a pre-approved
        # address means the user gets unsubscribed from the mailing list
        # without any further confirmation or approvals.
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne,
                                          pre_approved=True)
        list(workflow)
        # Anne is now unsubscribed form the mailing list.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNone(member)
        # No further token is needed.
        self.assertIsNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.no_one)

    def test_do_unsubscription_pre_approved_pre_confirmed(self):
        # A moderation-requiring unsubscription policy plus a pre-appvoed
        # address means the user gets unsubscribed to the mailing list without
        # any further confirmations or approvals.
        self._mlist.unsubscription_policy = (
            SubscriptionPolicy.confirm_then_moderate)
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne,
                                          pre_approved=True,
                                          pre_confirmed=True)
        list(workflow)
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNone(member)
        # No further token is needed.
        self.assertIsNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.no_one)

    def test_do_unsubscription_cleanups(self):
        # Once the user is unsubscribed, the token and its associated pending
        # database record will be removed from the database.
        self._mlist.unsubscription_policy = SubscriptionPolicy.open
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne,
                                          pre_approved=True,
                                          pre_confirmed=True)
        # Run the workflow.
        list(workflow)
        # Anne is now unsubscribed from the list.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNone(member)
        # Workflow is done, so it has no token.
        self.assertIsNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.no_one)

    def test_moderator_approves(self):
        # The workflow runs until moderator approval is required, at which
        # point the workflow is saved.  Once the moderator approves, the
        # workflow resumes and the user is unsubscribed.
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        workflow = UnSubscriptionWorkflow(
            self._mlist, self.anne, pre_confirmed=True)
        # Run the entire workflow.
        list(workflow)
        # The user is currently subscribed to the mailing list.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNotNone(member)
        self.assertIsNotNone(workflow.member)
        # The token is owned by the moderator.
        self.assertIsNotNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.moderator)
        # Create a new workflow with the previous workflow's save token, and
        # restore its state.  This models an approved un-sunscription request
        # and should result in the user getting subscribed.
        approved_workflow = UnSubscriptionWorkflow(self._mlist)
        approved_workflow.token = workflow.token
        approved_workflow.restore()
        list(approved_workflow)
        # Now the user is unsubscribed from the mailing list.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNone(member)
        self.assertEqual(approved_workflow.member, member)
        # No further token is needed.
        self.assertIsNone(approved_workflow.token)
        self.assertEqual(approved_workflow.token_owner, TokenOwner.no_one)

    def test_get_moderator_approval_log_on_hold(self):
        # When the unsubscription is held for moderator approval, a message is
        # logged.
        mark = LogFileMark('mailman.subscribe')
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        workflow = UnSubscriptionWorkflow(
            self._mlist, self.anne, pre_confirmed=True)
        # Run the entire workflow.
        list(workflow)
        self.assertIn(
         'test@example.com: held unsubscription request from anne@example.com',
         mark.readline()
         )
        # The state machine stopped at the moderator approval step so there
        # will be one token still in the database.
        self._expected_pendings_count = 1

    def test_get_moderator_approval_notifies_moderators(self):
        # When the unsubscription is held for moderator approval, and the list
        # is so configured, a notification is sent to the list moderators.
        self._mlist.admin_immed_notify = True
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        workflow = UnSubscriptionWorkflow(
            self._mlist, self.anne, pre_confirmed=True)
        # Consume the entire state machine.
        list(workflow)
        items = get_queue_messages('virgin', expected_count=1)
        message = items[0].msg
        self.assertEqual(message['From'], 'test-owner@example.com')
        self.assertEqual(message['To'], 'test-owner@example.com')
        self.assertEqual(
            message['Subject'],
            'New unsubscription request to Test from anne@example.com')
        self.assertEqual(message.get_payload(decode=True).decode('utf-8'), """\
Your authorization is required for a mailing list unsubscription
request approval:

    For:  Ане <anne@example.com>
    List: test@example.com
""")
        # The state machine stopped at the moderator approval so there will be
        # one token still in the database.
        self._expected_pendings_count = 1

    def test_get_moderator_approval_no_notifications(self):
        # When the unsubscription request is held for moderator approval, and
        # the list is so configured, a notification is sent to the list
        # moderators.
        self._mlist.admin_immed_notify = False
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        workflow = UnSubscriptionWorkflow(
            self._mlist, self.anne, pre_confirmed=True)
        # Consume the entire state machine.
        list(workflow)
        get_queue_messages('virgin', expected_count=0)
        # The state machine stopped at the moderator approval so there will be
        # one token still in the database.
        self._expected_pendings_count = 1

    def test_send_confirmation(self):
        # A confirmation message gets sent when the unsubscription must be
        # confirmed.
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        # Run the workflow to model the confirmation step.
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        list(workflow)
        items = get_queue_messages('virgin', expected_count=1)
        message = items[0].msg
        token = workflow.token
        self.assertTrue(
            str(message['Subject']).startswith('Your confirmation is '
                                               'needed to leave the '))
        self.assertEqual(
            message['From'], 'test-confirm+{}@example.com'.format(token))
        # The confirmation message is not `Precedence: bulk`.
        self.assertIsNone(message['precedence'])
        # The confirmation message is `Auto-Submitted: auto-generated`.
        self.assertEqual(message['auto-submitted'], 'auto-generated')
        # The state machine stopped at the member confirmation step so there
        # will be one token still in the database.
        self._expected_pendings_count = 1

    def test_do_confirmation_unsubscribes_user(self):
        # Unsubscriptions to the mailing list must be confirmed.  Once that's
        # done, the user's address is unsubscribed.
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        list(workflow)
        # Anne is a member.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNotNone(member)
        self.assertEqual(member, workflow.member)
        # The token is owned by the subscriber.
        self.assertIsNotNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.subscriber)
        # Confirm.
        confirm_workflow = UnSubscriptionWorkflow(self._mlist)
        confirm_workflow.token = workflow.token
        confirm_workflow.restore()
        list(confirm_workflow)
        # Anne is now unsubscribed.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNone(member)
        # No further token is needed.
        self.assertIsNone(confirm_workflow.token)
        self.assertEqual(confirm_workflow.token_owner, TokenOwner.no_one)

    def test_do_confirmation_unsubscribes_address(self):
        # Unsubscriptions to the mailing list must be confirmed.  Once that's
        # done, the address is unsubscribed.
        address = self.anne.register('anne.person@example.com')
        self._mlist.subscribe(address)
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        workflow = UnSubscriptionWorkflow(self._mlist, address)
        list(workflow)
        # Bart is a member.
        member = self._mlist.regular_members.get_member(
            'anne.person@example.com')
        self.assertIsNotNone(member)
        self.assertEqual(member, workflow.member)
        # The token is owned by the subscriber.
        self.assertIsNotNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.subscriber)
        # Confirm.
        confirm_workflow = UnSubscriptionWorkflow(self._mlist)
        confirm_workflow.token = workflow.token
        confirm_workflow.restore()
        list(confirm_workflow)
        # Bart is now unsubscribed.
        member = self._mlist.regular_members.get_member(
            'anne.person@example.com')
        self.assertIsNone(member)
        # No further token is needed.
        self.assertIsNone(confirm_workflow.token)
        self.assertEqual(confirm_workflow.token_owner, TokenOwner.no_one)

    def test_do_confirmation_nonmember(self):
        # Attempt to confirm the unsubscription of a member who has already
        # been unsubscribed.
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        list(workflow)
        # Anne is a member.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNotNone(member)
        self.assertEqual(member, workflow.member)
        # The token is owned by the subscriber.
        self.assertIsNotNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.subscriber)
        # Unsubscribe Anne out of band.
        member.unsubscribe()
        # Confirm.
        confirm_workflow = UnSubscriptionWorkflow(self._mlist)
        confirm_workflow.token = workflow.token
        confirm_workflow.restore()
        list(confirm_workflow)
        # No further token is needed.
        self.assertIsNone(confirm_workflow.token)
        self.assertEqual(confirm_workflow.token_owner, TokenOwner.no_one)

    def test_do_confirmation_nonmember_final_step(self):
        # Attempt to confirm the unsubscription of a member who has already
        # been unsubscribed.
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        list(workflow)
        # Anne is a member.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNotNone(member)
        self.assertEqual(member, workflow.member)
        # The token is owned by the subscriber.
        self.assertIsNotNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.subscriber)
        # Confirm.
        confirm_workflow = UnSubscriptionWorkflow(self._mlist)
        confirm_workflow.token = workflow.token
        confirm_workflow.restore()
        confirm_workflow.run_until('do_unsubscription')
        self.assertEqual(member, confirm_workflow.member)
        # Unsubscribe Anne out of band.
        member.unsubscribe()
        list(confirm_workflow)
        self.assertIsNone(confirm_workflow.member)
        # No further token is needed.
        self.assertIsNone(confirm_workflow.token)
        self.assertEqual(confirm_workflow.token_owner, TokenOwner.no_one)

    def test_prevent_confirmation_replay_attacks(self):
        # Ensure that if the workflow requires two confirmations, e.g. first
        # the user confirming their subscription, and then the moderator
        # approving it, that different tokens are used in these two cases.
        self._mlist.unsubscription_policy = (
            SubscriptionPolicy.confirm_then_moderate)
        workflow = UnSubscriptionWorkflow(self._mlist, self.anne)
        # Run the state machine up to the first confirmation, and cache the
        # confirmation token.
        list(workflow)
        token = workflow.token
        # Anne is still a member of the mailing list.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNotNone(member)
        self.assertIsNotNone(workflow.member)
        # The token is owned by the subscriber.
        self.assertIsNotNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.subscriber)
        # The old token will not work for moderator approval.
        moderator_workflow = UnSubscriptionWorkflow(self._mlist)
        moderator_workflow.token = token
        moderator_workflow.restore()
        list(moderator_workflow)
        # The token is owned by the moderator.
        self.assertIsNotNone(moderator_workflow.token)
        self.assertEqual(moderator_workflow.token_owner, TokenOwner.moderator)
        # While we wait for the moderator to approve the subscription, note
        # that there's a new token for the next steps.
        self.assertNotEqual(token, moderator_workflow.token)
        # The old token won't work.
        final_workflow = UnSubscriptionWorkflow(self._mlist)
        final_workflow.token = token
        self.assertRaises(LookupError, final_workflow.restore)
        # Running this workflow will fail.
        self.assertRaises(AssertionError, list, final_workflow)
        # Anne is still not unsubscribed.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNotNone(member)
        self.assertIsNone(final_workflow.member)
        # However, if we use the new token, her unsubscription request will be
        # approved by the moderator.
        final_workflow.token = moderator_workflow.token
        final_workflow.restore()
        list(final_workflow)
        # And now Anne is unsubscribed.
        member = self._mlist.regular_members.get_member(self._anne)
        self.assertIsNone(member)
        # No further token is needed.
        self.assertIsNone(final_workflow.token)
        self.assertEqual(final_workflow.token_owner, TokenOwner.no_one)

    def test_confirmation_needed_and_pre_confirmed(self):
        # The subscription policy is 'confirm' but the subscription is
        # pre-confirmed so the moderation checks can be skipped.
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        workflow = UnSubscriptionWorkflow(
            self._mlist, self.anne, pre_confirmed=True, pre_approved=True)
        list(workflow)
        # Anne was unsubscribed.
        self.assertIsNone(workflow.token)
        self.assertEqual(workflow.token_owner, TokenOwner.no_one)
        self.assertIsNone(workflow.member)

    def test_confirmation_needed_moderator_address(self):
        address = self.anne.register('anne.person@example.com')
        self._mlist.subscribe(address)
        self._mlist.unsubscription_policy = SubscriptionPolicy.moderate
        workflow = UnSubscriptionWorkflow(self._mlist, address)
        # Get moderator approval.
        list(workflow)
        approved_workflow = UnSubscriptionWorkflow(self._mlist)
        approved_workflow.token = workflow.token
        approved_workflow.restore()
        list(approved_workflow)
        self.assertEqual(approved_workflow.subscriber, address)
        # Anne was unsubscribed.
        self.assertIsNone(approved_workflow.token)
        self.assertEqual(approved_workflow.token_owner, TokenOwner.no_one)
        self.assertIsNone(approved_workflow.member)
        member = self._mlist.regular_members.get_member(
            'anne.person@example.com')
        self.assertIsNone(member)



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Migration from Python 2 to Python 3.

Some columns changed from LargeBinary type to Unicode type.

Revision ID: 33e1f5f6fa8
Revises: 51b7f92bd06c
Create Date: 2015-01-20 17:32:30.144083

"""

import sqlalchemy as sa

from alembic import op
from mailman.database.helpers import is_sqlite
from mailman.database.types import SAUnicode


# Revision identifiers, used by Alembic.
revision = '33e1f5f6fa8'
down_revision = '51b7f92bd06c'


COLUMNS_TO_CHANGE = (
    ('message', 'message_id_hash'),
    ('message', 'path'),
    ('pended', 'token'),
    ('_request', 'data_hash'),
    ('user', 'password'),
    )


def upgrade():
    if is_sqlite(op.get_bind()):
        # SQLite does not support altering columns.
        return
    for table, column in COLUMNS_TO_CHANGE:
        op.alter_column(table, column, type_=SAUnicode)


def downgrade():
    if is_sqlite(op.get_bind()):
        # SQLite does not support altering columns.
        return
    for table, column in COLUMNS_TO_CHANGE:
        if op.get_bind().dialect.name == 'postgresql':
            # PostgreSQL needs the USING clause that Alembic does not support
            # yet.
            op.execute(
                ('ALTER TABLE "{table}" ALTER COLUMN "{column}" '
                 'TYPE BYTEA USING decode("{column}", \'UTF8\')').format(
                     table=table, column=column))
        else:
            op.alter_column(table, column, type_=sa.LargeBinary)



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""REST root object tests."""

import os
import requests
import unittest

from mailman.config import config
from mailman.core.system import system
from mailman.database.transaction import transaction
from mailman.interfaces.template import ITemplateManager
from mailman.testing.helpers import call_api
from mailman.testing.layers import RESTLayer
from urllib.error import HTTPError
from zope.component import getUtility


class TestRoot(unittest.TestCase):
    layer = RESTLayer

    def test_root_system_backward_compatibility(self):
        # The deprecated path for getting system version information points
        # you to the new URL.
        url = 'http://localhost:9001/3.0/system'
        new = '{}/versions'.format(url)
        json, response = call_api(url)
        self.assertEqual(json['mailman_version'], system.mailman_version)
        self.assertEqual(json['python_version'], system.python_version)
        self.assertEqual(json['self_link'], new)

    def test_system_versions(self):
        # System version information is available via REST.
        url = 'http://localhost:9001/3.0/system/versions'
        json, response = call_api(url)
        self.assertEqual(json['mailman_version'], system.mailman_version)
        self.assertEqual(json['python_version'], system.python_version)
        self.assertEqual(json['self_link'], url)

    def test_path_under_root_does_not_exist(self):
        # Accessing a non-existent path under root returns a 404.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/does-not-exist')
        self.assertEqual(cm.exception.code, 404)

    def test_system_url_not_preferences(self):
        # /system/foo where `foo` is not `preferences`.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/foo')
        self.assertEqual(cm.exception.code, 404)

    def test_system_preferences_are_read_only(self):
        # /system/preferences are read-only.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/preferences', {
                     'acknowledge_posts': True,
                     }, method='PATCH')
        self.assertEqual(cm.exception.code, 405)
        # /system/preferences are read-only.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/preferences', {
                'acknowledge_posts': False,
                'delivery_mode': 'regular',
                'delivery_status': 'enabled',
                'hide_address': True,
                'preferred_language': 'en',
                'receive_list_copy': True,
                'receive_own_postings': True,
                }, method='PUT')
        self.assertEqual(cm.exception.code, 405)

    def test_queue_directory(self):
        # The REST runner is not queue runner, so it should not have a
        # directory in var/queue.
        queue_directory = os.path.join(config.QUEUE_DIR, 'rest')
        self.assertFalse(os.path.isdir(queue_directory))

    def test_no_basic_auth(self):
        # If Basic Auth credentials are missing, it is a 401 error.
        response = requests.get('http://localhost:9001/3.0/system')
        self.assertEqual(response.status_code, 401)
        json = response.json()
        self.assertEqual(json['title'], '401 Unauthorized')
        self.assertEqual(json['description'], 'REST API authorization failed')

    def test_unauthorized(self):
        # Bad Basic Auth credentials results in a 401 error.
        response = requests.get(
            'http://localhost:9001/3.0/system',
            auth=('baduser', 'badpass'))
        self.assertEqual(response.status_code, 401)
        json = response.json()
        self.assertEqual(json['title'], '401 Unauthorized')
        self.assertEqual(json['description'], 'REST API authorization failed')

    def test_reserved_bad_subpath(self):
        # Only <api>/reserved/uids/orphans is a defined resource.  DELETEing
        # anything else gives a 404.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/reserved/uids/assigned',
                     method='DELETE')
        self.assertEqual(cm.exception.code, 404)

    def test_system_pipelines_are_exposed(self):
        json, response = call_api('http://localhost:9001/3.0/system/pipelines')
        self.assertEqual(json['pipelines'], sorted(config.pipelines))

    def test_system_pipelines_bad_request(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/pipelines/bogus')
        self.assertEqual(cm.exception.code, 400)

    def test_system_pipelines_are_read_only(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/pipelines', {
                'pipelines': ['pipeline_1', 'pipeline_2']
                }, method='PATCH')
        self.assertEqual(cm.exception.code, 405)
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/pipelines', {
                'pipelines': ['pipeline_1', 'pipeline_2']
                }, method='PUT')
        self.assertEqual(cm.exception.code, 405)

    def test_system_chains_are_exposed(self):
        json, response = call_api('http://localhost:9001/3.0/system/chains')
        self.assertEqual(json['chains'], sorted(config.chains))

    def test_system_chains_bad_request(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/chains/bogus')
        self.assertEqual(cm.exception.code, 400)

    def test_system_chains_are_read_only(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/chains', {
                'chains': ['chain_1', 'chain_2']
                }, method='PATCH')
        self.assertEqual(cm.exception.code, 405)
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/system/chains', {
                'chains': ['chain_1', 'chain_2']
                }, method='PUT')
        self.assertEqual(cm.exception.code, 405)


class TestSiteTemplates(unittest.TestCase):
    """Test /uris"""

    layer = RESTLayer

    def test_no_templates_for_api_30(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/uris')
        self.assertEqual(cm.exception.code, 404)

    def test_path_too_long(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/uris/foo/bar')
        self.assertEqual(cm.exception.code, 400)

    def test_get_unknown_uri(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/uris/not:a:template')
        self.assertEqual(cm.exception.code, 404)

    def test_get_all_uris(self):
        manager = getUtility(ITemplateManager)
        with transaction():
            manager.set(
                'list:user:notice:welcome', None,
                'http://example.com/welcome')
            manager.set(
                'list:user:notice:goodbye', None,
                'http://example.com/goodbye',
                'a user', 'the password',
                )
        json, response = call_api(
            'http://localhost:9001/3.1/uris')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json['start'], 0)
        self.assertEqual(json['total_size'], 2)
        self.assertEqual(
            json['self_link'],
            'http://localhost:9001/3.1/uris')
        self.assertEqual(json['entries'], [
            {'http_etag': '"063fd6635a6035a4b7e939a304fcbd16571aa662"',
             'name': 'list:user:notice:goodbye',
             'password': 'the password',
             'self_link': ('http://localhost:9001/3.1'
                           '/uris/list:user:notice:goodbye'),
             'uri': 'http://example.com/goodbye',
             'username': 'a user',
             },
            {'http_etag': '"5c4ec63b2a0a50f96483ec85b94b80ee092af792"',
             'name': 'list:user:notice:welcome',
             'self_link': ('http://localhost:9001/3.1'
                           '/uris/list:user:notice:welcome'),
             'uri': 'http://example.com/welcome',
             }])

    def test_patch_uris(self):
        json, response = call_api(
            'http://localhost:9001/3.1/uris', {
                'list:user:notice:welcome': 'http://example.org/welcome',
                'list:user:notice:goodbye': 'http://example.org/goodbye',
                }, method='PATCH')
        self.assertEqual(response.status_code, 204)
        manager = getUtility(ITemplateManager)
        template = manager.raw('list:user:notice:welcome', None)
        self.assertEqual(template.uri, 'http://example.org/welcome')
        self.assertIsNone(template.username)
        self.assertEqual(template.password, '')
        template = manager.raw('list:user:notice:goodbye', None)
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertIsNone(template.username)
        self.assertEqual(template.password, '')

    def test_patch_uris_with_credentials(self):
        json, response = call_api(
            'http://localhost:9001/3.1/uris', {
                'list:user:notice:welcome': 'http://example.org/welcome',
                'list:user:notice:goodbye': 'http://example.org/goodbye',
                'password': 'some password',
                'username': 'anne.person',
                }, method='PATCH')
        self.assertEqual(response.status_code, 204)
        manager = getUtility(ITemplateManager)
        template = manager.raw('list:user:notice:welcome', None)
        self.assertEqual(template.uri, 'http://example.org/welcome')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:user:notice:goodbye', None)
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')

    def test_patch_uris_with_partial_credentials(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/uris', {
                    'list:user:notice:welcome': 'http://example.org/welcome',
                    'list:user:notice:goodbye': 'http://example.org/goodbye',
                    'username': 'anne.person',
                    }, method='PATCH')
        self.assertEqual(cm.exception.code, 400)

    def test_put_all_uris(self):
        json, response = call_api(
            'http://localhost:9001/3.1/uris', {
                'domain:admin:notice:new-list': '',
                'list:admin:action:post': '',
                'list:admin:action:subscribe': '',
                'list:admin:action:unsubscribe': '',
                'list:admin:notice:subscribe': '',
                'list:admin:notice:unrecognized': '',
                'list:admin:notice:unsubscribe': '',
                'list:admin:notice:pending': '',
                'list:admin:notice:increment': '',
                'list:admin:notice:disable': '',
                'list:admin:notice:removal': '',
                'list:member:digest:footer': '',
                'list:member:digest:header': '',
                'list:member:digest:masthead': '',
                'list:member:regular:footer': 'http://example.org/footer',
                'list:member:regular:header': 'http://example.org/header',
                'list:user:action:invite': '',
                'list:user:action:subscribe': '',
                'list:user:action:unsubscribe': '',
                'list:user:notice:goodbye': 'http://example.org/goodbye',
                'list:user:notice:hold': '',
                'list:user:notice:no-more-today': '',
                'list:user:notice:post': '',
                'list:user:notice:probe': '',
                'list:user:notice:refuse': '',
                'list:user:notice:rejected': '',
                'list:user:notice:welcome': 'http://example.org/welcome',
                'list:user:notice:warning': '',
                'password': 'some password',
                'username': 'anne.person',
                }, method='PUT')
        self.assertEqual(response.status_code, 204)
        manager = getUtility(ITemplateManager)
        template = manager.raw('list:member:digest:footer', None)
        self.assertIsNone(template)
        template = manager.raw('list:member:digest:header', None)
        self.assertIsNone(template)
        template = manager.raw('list:member:regular:footer', None)
        self.assertEqual(template.uri, 'http://example.org/footer')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:member:regular:header', None)
        self.assertEqual(template.uri, 'http://example.org/header')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:user:notice:goodbye', None)
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:user:notice:goodbye', None)
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')

    def test_delete_all_uris(self):
        manager = getUtility(ITemplateManager)
        with transaction():
            manager.set(
                'list:user:notice:welcome', None,
                'http://example.com/welcome')
            manager.set(
                'list:user:notice:goodbye', None,
                'http://example.com/goodbye',
                'a user', 'the password',
                )
        json, response = call_api(
            'http://localhost:9001/3.1/uris',
            method='DELETE')
        self.assertEqual(response.status_code, 204)
        self.assertIsNone(manager.raw('list:user:notice:welcome', None))
        self.assertIsNone(manager.raw('list:user:notice:goodbye', None))

    def test_get_a_url(self):
        with transaction():
            getUtility(ITemplateManager).set(
                'list:user:notice:welcome', None,
                'http://example.com/welcome')
        json, response = call_api(
            'http://localhost:9001/3.1/uris/list:user:notice:welcome')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json, {
            'http_etag': '"86e360d83197561d50826ad6d15e9c30923b82d6"',
            'self_link': ('http://localhost:9001/3.1'
                          '/uris/list:user:notice:welcome'),
            'uri': 'http://example.com/welcome',
            })

    def test_get_a_bad_url(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/uris/list:user:notice:notemplate')
        self.assertEqual(cm.exception.code, 404)

    def test_get_unset_url(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/uris/list:user:notice:welcome')
        self.assertEqual(cm.exception.code, 404)

    def test_patch_url_with_too_many_parameters(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/uris', {
                    'list:user:notice:welcome': 'http://example.org/welcome',
                    'list:user:notice:goodbye': 'http://example.org/goodbye',
                    'secret': 'some password',
                    'person': 'anne.person',
                    }, method='PATCH')
        self.assertEqual(cm.exception.code, 400)



# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test for the DMARC chain."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.core.chains import process as process_chain
from mailman.interfaces.chain import DiscardEvent, RejectEvent
from mailman.testing.helpers import (
    event_subscribers,
    get_queue_messages,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer


class TestDMARC(unittest.TestCase):
    layer = ConfigLayer
    maxDiff = None

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        self._msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: Ignore

""")

    def test_discard(self):
        msgdata = dict(dmarc_action='discard')
        # When a message is discarded, the only artifacts are a log message
        # and an event.  Catch the event to prove it happened.
        events = []
        def handler(event):                         # noqa: E306
            if isinstance(event, DiscardEvent):
                events.append(event)
        with event_subscribers(handler):
            process_chain(self._mlist, self._msg, msgdata, start_chain='dmarc')
        self.assertEqual(len(events), 1)
        self.assertIs(events[0].msg, self._msg)

    def test_reject(self):
        msgdata = dict(
            dmarc_action='reject',
            moderation_reasons=['DMARC violation'],
            )
        # When a message is reject, an event will be triggered and the message
        # will be bounced.
        events = []
        def handler(event):                         # noqa: E306
            if isinstance(event, RejectEvent):
                events.append(event)
        with event_subscribers(handler):
            process_chain(self._mlist, self._msg, msgdata, start_chain='dmarc')
        self.assertEqual(len(events), 1)
        self.assertIs(events[0].msg, self._msg)
        items = get_queue_messages('virgin', expected_count=1)
        # Unpack the rejection message.
        rejection = items[0].msg.get_payload(0).get_payload()
        self.assertEqual(rejection, """\
Your message to the Ant mailing-list was rejected for the following
reasons:

DMARC violation

The original message as received by Mailman is attached.
""")



# Copyright (C) 2008-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interfaces defining email commands."""

from enum import Enum
from public import public
from zope.interface import Attribute, Interface


@public
class ContinueProcessing(Enum):
    """Should `IEmailCommand.process()` continue or not."""
    no = 0
    yes = 1


@public
class IEmailResults(Interface):
    """The email command results object."""

    output = Attribute('An output file object for printing results to.')


@public
class IEmailCommand(Interface):
    """An email command."""

    name = Attribute('Command name as seen in a -request email.')

    argument_description = Attribute('Description of command arguments.')

    description = Attribute('Command help.')

    def process(mlist, msg, msgdata, arguments, results):
        """Process the email command.

        :param mlist: The mailing list target of the command.
        :param msg: The original message object.
        :param msgdata: The message metadata.
        :param arguments: The command arguments tuple.
        :param results: An IEmailResults object for these commands.
        :return: A `ContinueProcessing` enum specifying whether to continue
            processing or not.
        """


@public
class ICLISubCommand(Interface):
    """A command line interface subcommand.

    Subcommands are implemented using the `click` package.  See
    https://click.palletsprojects.com/en/7.x/ for details.
    """
    name = Attribute(
        """The subcommand name as it will show up in `mailman --help`.

        This must be unique; it is a runtime error if any plugin provides a
        subcommand with a clashing name.
        """)

    command = Attribute(
        """The click command to run for this subcommand.

        This must be a function decorated with at least the @click.command()
        decorator.  The function may also be decorated with other arguments as
        needed.
        """)



"""Pendable indexes

Add indexes on Pendable fields that can be queried upon.


Revision ID: 47294d3a604
Revises: 33bc0099223
Create Date: 2015-12-02 11:46:47.295174

"""

import json
import sqlalchemy as sa

from alembic import op
from mailman.database.types import SAUnicode


# revision identifiers, used by Alembic.
revision = '47294d3a604'
down_revision = '3e09bb4a5dc'


TYPE_CLUES = {
    'member_id': 'probe',
    'token_owner': 'subscription',
    '_mod_message_id': 'data',
    }

pended_table = sa.sql.table(
    'pended',
    sa.sql.column('id', sa.Integer),
    )

keyvalue_table = sa.sql.table(
    'pendedkeyvalue',
    sa.sql.column('id', sa.Integer),
    sa.sql.column('key', SAUnicode),
    sa.sql.column('value', SAUnicode),
    sa.sql.column('pended_id', sa.Integer),
    )


def upgrade():
    op.create_index(
        op.f('ix_pended_expiration_date'), 'pended', ['expiration_date'],
        unique=False)
    op.create_index(op.f('ix_pended_token'), 'pended', ['token'], unique=False)
    op.create_index(
        op.f('ix_pendedkeyvalue_key'), 'pendedkeyvalue', ['key'], unique=False)
    op.create_index(
        op.f('ix_pendedkeyvalue_value'), 'pendedkeyvalue', ['value'],
        unique=False)
    # Data migration.
    connection = op.get_bind()
    for pended_id in connection.execute(pended_table.select()).scalars().fetchall():  # noqa: E501
        keyvalues = connection.execute(keyvalue_table.select().where(
            keyvalue_table.c.pended_id == pended_id
            )).fetchall()
        kv_type = [kv for kv in keyvalues if kv[1] == 'type']
        if kv_type:
            # Convert existing type keys from JSON to plain text.
            # The (pended_id, key) tuple is unique.
            kv_type = kv_type[0]
            try:
                new_value = json.loads(kv_type[2])
            except ValueError:
                # New-style entry (or already converted).
                pass
            else:
                connection.execute(keyvalue_table.update().where(
                    keyvalue_table.c.id == kv_type[0]
                    ).values(value=new_value))
        else:
            # Detect the type and add the corresponding type key.
            keys = [kv[1] for kv in keyvalues]
            for clue_key, clue_type in TYPE_CLUES.items():
                if clue_key not in keys:
                    continue
                # We found the type, update the DB.
                connection.execute(keyvalue_table.insert().values(
                    key='type', value=clue_type, pended_id=pended_id))
                break


def downgrade():
    # Data migration.
    connection = op.get_bind()
    # Remove the introduced type keys.
    connection.execute(keyvalue_table.delete().where(sa.and_(
        keyvalue_table.c.key == 'type',
        keyvalue_table.c.value.in_(TYPE_CLUES.values())
        )))
    # Convert the other type keys to JSON.
    keyvalues = connection.execute(sa.select(keyvalue_table.c.id, keyvalue_table.c.value).where(  # noqa: E501
        keyvalue_table.c.key == 'type')).fetchall()
    for keyvalue in keyvalues:
        connection.execute(keyvalue_table.update().where(
            keyvalue_table.c.id == keyvalue[0]
            ).values(value=json.dumps(keyvalue[1])))
    # Remove indexes.
    op.drop_index(op.f('ix_pendedkeyvalue_value'), table_name='pendedkeyvalue')
    op.drop_index(op.f('ix_pendedkeyvalue_key'), table_name='pendedkeyvalue')
    op.drop_index(op.f('ix_pended_token'), table_name='pended')
    op.drop_index(op.f('ix_pended_expiration_date'), table_name='pended')



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the owner chain."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.chains.owner import BuiltInOwnerChain
from mailman.core.chains import process
from mailman.interfaces.chain import AcceptOwnerEvent
from mailman.testing.helpers import (
    event_subscribers,
    get_queue_messages,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import ConfigLayer


class TestOwnerChain(unittest.TestCase):
    """Test the owner chain."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._msg = mfs("""\
From: anne@example.com
To: test@example.com
Message-ID: <ant>

""")

    def test_owner_pipeline(self):
        # Messages processed through the default owners chain end up in the
        # pipeline queue, and an event gets sent.
        #
        # This event subscriber records the event that occurs when the message
        # is processed by the owner chain.
        events = []
        def catch_event(event):                                  # noqa: E306
            if isinstance(event, AcceptOwnerEvent):
                events.append(event)
        with event_subscribers(catch_event):
            process(self._mlist, self._msg, {}, 'default-owner-chain')
        self.assertEqual(len(events), 1)
        event = events[0]
        self.assertIsInstance(event, AcceptOwnerEvent)
        self.assertEqual(event.mlist, self._mlist)
        self.assertEqual(event.msg['message-id'], '<ant>')
        self.assertIsInstance(event.chain, BuiltInOwnerChain)
        items = get_queue_messages('pipeline', expected_count=1)
        message = items[0].msg
        self.assertEqual(message['message-id'], '<ant>')



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Filesystem utilities."""

import os
import sys

from contextlib import suppress
from public import public


@public
class umask:
    """Manage the umask for the with statement."""

    def __init__(self, umask):
        self._umask = umask
        self._old_umask = None

    def __enter__(self):
        assert self._old_umask is None, 'Unexpected existing umask'
        self._old_umask = os.umask(self._umask)

    def __exit__(self, *exc_info):
        assert self._old_umask is not None, 'No previous umask'
        os.umask(self._old_umask)
        # Do not suppress exceptions.
        return False


@public
def makedirs(path, mode=0o0755):
    """Create a directory hierarchy, ensuring permissions.

    Other than just calling os.makedirs(), this ensures that the umask is
    reset so that the makedirs mode will be honored.

    :param path: The directory path to create.
    :type path: string
    :param mode: The numeric permission mode to use.
    :type mode: int
    """
    with umask(0):
        # In order for os.walk to set permissions appropriately if required,
        # the FIRST NON EXISTENT parent directory of the one we wanted to
        # create has to be provided.
        upwards = first_inexistent_directory(path)

        # The directory exists, nothing to do.
        if upwards is None:
            return

        try:
            os.makedirs(path, mode)
        except FileExistsError:
            if not os.path.isdir(path):
                raise FileExistsError((
                   "A race condition might have happened. {} actually "
                   "exists and is not a directory.").format(path))

    # Some systems such as FreeBSD ignore mkdir's mode, so walk the just
    # created directories and try to set the mode, ignoring any OSErrors that
    # occur here.
    for dirpath, dirnames, filenames in os.walk(upwards):
        with suppress(OSError):
            os.chmod(dirpath, mode)


@public
def safe_remove(path):
    with suppress(FileNotFoundError):
        os.remove(path)


@public
def first_inexistent_directory(path):
    """Splits iteratively a path until it gives the first non-existent
    directory in the tree.

    That is, if /home/user/foo/bar/baz is given to the function, and
    /home/user/foo/bar doesn't exist but /home/user/foo exists, returns
    /home/user/foo/bar.
    """
    directory = path
    rhs = None

    while True:
        if os.path.isdir(directory):
            if rhs is None:
                return None
            else:
                return os.path.join(directory, rhs)
        elif os.path.exists(directory):
            raise FileExistsError(
                "The path %s exists but is not a directory.",
                directory)
        directory, rhs = os.path.split(directory)


@public
def path(package, module, *args, **kw):
    """Wrap around importlib.resources.path.

    importlib_resources.path (PyPI package we use for compatibility in Python <
    3.7) has now diverged in behavior from importlib.resources.path (in Python
    >= 3.7), especially in terms of supporting directories. Even though we can
    just jump to the new version of the library, many distributions packaging
    Mailman do not package importlib_resources at all and instead patch the
    source code to simply replace importlib_resources with importlib.resources.

    This utility method is meant to keep that patching ability without any
    complicated patches to make Mailman work with standard library
    importlib.resources. This is only supposed to be used where the divergent
    behavior causes problems for us.
    """
    # Note to packaging teams: This function will handle both standard library
    # and 3rd party importlib_resources package. Please do not patch it.
    if module:
        module_package = '{}.{}'.format(package, module)
    else:
        module_package = package

    try:
        if sys.version_info < (3, 9):
            from importlib.resources import path
            return path(package, module, *args, **kw)
        else:
            from importlib.resources import files  # pragma: nocover
            return files(module_package, *args, **kw)         # pragma: nocover
    except ImportError:                                       # pragma: nocover
        from importlib_resources import files  # pragma: nocover
        return files(module_package, *args, **kw)             # pragma: nocover
